# -*- coding: utf-8 -*-

"""Cross references from Gilda.

.. seealso:: https://github.com/indralabs/gilda
"""

import bioregistry
import pandas as pd

from pyobo.constants import (
    PROVENANCE,
    SOURCE_ID,
    SOURCE_PREFIX,
    TARGET_ID,
    TARGET_PREFIX,
)

__all__ = [
    "get_gilda_xrefs_df",
]

GILDA_MAPPINGS = (
    "https://raw.githubusercontent.com/indralab/gilda/master/gilda/resources/mesh_mappings.tsv"
)


def get_gilda_xrefs_df() -> pd.DataFrame:
    """Get xrefs from Gilda."""
    df = pd.read_csv(
        GILDA_MAPPINGS,
        sep="\t",
        header=None,
        usecols=[0, 1, 3, 4],
        names=[SOURCE_PREFIX, SOURCE_ID, TARGET_PREFIX, TARGET_ID],
    )
    df[PROVENANCE] = GILDA_MAPPINGS

    for k in SOURCE_PREFIX, TARGET_PREFIX:
        df[k] = df[k].map(bioregistry.normalize_prefix)

    for k in SOURCE_ID, TARGET_ID:
        df[k] = df[k].map(_fix_gogo)

    return df


def _fix_gogo(s):
    for prefix in ("CHEBI:", "DOID:", "HP:", "GO:"):
        if s.startswith(prefix):
            return s[len(prefix) :]
    return s



# -*- coding: utf-8 -*-

"""Converter for HGNC Gene Families."""

from collections import defaultdict
from typing import Iterable, List, Mapping

import pandas as pd

from ..struct import (
    Obo,
    Reference,
    Synonym,
    SynonymTypeDef,
    Term,
    enables,
    from_species,
)
from ..utils.path import ensure_path

__all__ = [
    "HGNCGroupGetter",
]

PREFIX = "hgnc.genegroup"
FAMILIES_URL = "ftp://ftp.ebi.ac.uk/pub/databases/genenames/new/csv/genefamily_db_tables/family.csv"
# TODO use family_alias.csv
HIERARCHY_URL = (
    "ftp://ftp.ebi.ac.uk/pub/databases/genenames/new/csv/genefamily_db_tables/hierarchy.csv"
)

symbol_type = SynonymTypeDef.from_text("symbol")


class HGNCGroupGetter(Obo):
    """An ontology representation of HGNC's gene group nomenclature."""

    ontology = PREFIX
    dynamic_version = True
    synonym_typedefs = [symbol_type]
    typedefs = [from_species, enables]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force)


def get_obo(force: bool = False) -> Obo:
    """Get HGNC Gene Groups as OBO."""
    return HGNCGroupGetter(force=force)


def get_hierarchy(force: bool = False) -> Mapping[str, List[str]]:
    """Get the HGNC Gene Families hierarchy as a dictionary."""
    path = ensure_path(PREFIX, url=HIERARCHY_URL, force=force)
    df = pd.read_csv(path, dtype={"parent_fam_id": str, "child_fam_id": str})
    d = defaultdict(list)
    for parent_id, child_id in df.values:
        d[child_id].append(parent_id)
    return dict(d)


COLUMNS = ["id", "abbreviation", "name", "pubmed_ids", "desc_comment", "desc_go"]


def get_terms(force: bool = False) -> Iterable[Term]:
    """Get the HGNC Gene Group terms."""
    terms = list(_get_terms_helper(force=force))
    hierarchy = get_hierarchy(force=force)

    id_to_term = {term.reference.identifier: term for term in terms}
    for child_id, parent_ids in hierarchy.items():
        child: Term = id_to_term[child_id]
        for parent_id in parent_ids:
            parent: Term = id_to_term[parent_id]
            child.append_parent(
                Reference(
                    prefix=PREFIX,
                    identifier=parent_id,
                    name=parent.name,
                )
            )
    gene_group = Reference.auto("SO", "0005855")
    yield Term(reference=gene_group)
    for term in terms:
        if not term.parents:
            term.append_parent(gene_group)
    yield from terms


def _get_terms_helper(force: bool = False) -> Iterable[Term]:
    path = ensure_path(PREFIX, url=FAMILIES_URL, force=force)
    df = pd.read_csv(path, dtype={"id": str})

    for gene_group_id, symbol, name, pubmed_ids, definition, desc_go in df[COLUMNS].values:
        if not definition or pd.isna(definition):
            definition = None
        term = Term(
            reference=Reference(prefix=PREFIX, identifier=gene_group_id, name=name),
            definition=definition,
        )
        if pubmed_ids and pd.notna(pubmed_ids):
            for s in pubmed_ids.split(","):
                term.append_provenance(Reference(prefix="pubmed", identifier=s.strip()))
        if desc_go and pd.notna(desc_go):
            go_id = desc_go[len("http://purl.uniprot.org/go/") :]
            term.append_relationship(enables, Reference(prefix="GO", identifier=go_id))
        if symbol and pd.notna(symbol):
            term.append_synonym(Synonym(name=symbol, type=symbol_type))
        term.set_species(identifier="9606", name="Homo sapiens")
        yield term


if __name__ == "__main__":
    HGNCGroupGetter.cli()



# -*- coding: utf-8 -*-

"""A python package for handling and generating OBO."""

from .api import (  # noqa: F401
    get_alts_to_id,
    get_ancestors,
    get_definition,
    get_descendants,
    get_filtered_properties_df,
    get_filtered_properties_mapping,
    get_filtered_properties_multimapping,
    get_filtered_relations_df,
    get_filtered_xrefs,
    get_hierarchy,
    get_id_definition_mapping,
    get_id_multirelations_mapping,
    get_id_name_mapping,
    get_id_species_mapping,
    get_id_synonyms_mapping,
    get_id_to_alts,
    get_ids,
    get_name,
    get_name_by_curie,
    get_name_id_mapping,
    get_primary_curie,
    get_primary_identifier,
    get_properties,
    get_properties_df,
    get_property,
    get_relation,
    get_relation_mapping,
    get_relations_df,
    get_species,
    get_sssom_df,
    get_subhierarchy,
    get_synonyms,
    get_typedef_df,
    get_xref,
    get_xrefs,
    get_xrefs_df,
    has_ancestor,
    is_descendent,
)
from .getters import get_ontology  # noqa: F401
from .identifier_utils import normalize_curie  # noqa: F401
from .normalizer import OboNormalizer, ground  # noqa: F401
from .obographs import parse_results_from_obo  # noqa: F401
from .plugins import (  # noqa: F401
    has_nomenclature_plugin,
    iter_nomenclature_plugins,
    run_nomenclature_plugin,
)
from .reader import from_obo_path, from_obonet  # noqa: F401
from .struct import Obo, Reference, Synonym, SynonymTypeDef, Term, TypeDef  # noqa: F401
from .utils.path import ensure_path  # noqa: F401
from .version import get_version  # noqa: F401
from .xrefdb.canonicalizer import (  # noqa: F401
    Canonicalizer,
    get_equivalent,
    get_priority_curie,
)
from .xrefdb.sources import (  # noqa: F401
    has_xref_plugin,
    iter_xref_plugins,
    run_xref_plugin,
)



# -*- coding: utf-8 -*-

"""Converter for the Integrated Taxonomic Information System (ITIS)."""

import os
import shutil
import sqlite3
import zipfile
from contextlib import closing
from typing import Iterable

from pyobo.struct import Obo, Reference, Term
from pyobo.utils.io import multidict
from pyobo.utils.path import ensure_path, prefix_directory_join

__all__ = [
    "ITISGetter",
]

PREFIX = "itis"
URL = "https://www.itis.gov/downloads/itisSqlite.zip"

LONGNAMES_QUERY = """
SELECT tsn, completename
FROM longnames
"""

HIERARCHY_QUERY = """
SELECT TSN, Parent_TSN
FROM hierarchy
"""


# TODO confusing logic since you need to download the data first to get the version


class ITISGetter(Obo):
    """An ontology representation of the ITIS taxonomy."""

    ontology = bioversions_key = PREFIX

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        # don't add force since the version getter already will
        return iter_terms(force=force, version=self._version_or_raise)


def get_obo() -> Obo:
    """Get ITIS as OBO."""
    return ITISGetter()


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get ITIS terms."""
    zip_path = ensure_path(PREFIX, url=URL, force=force, version=version)
    sqlite_dir = prefix_directory_join(PREFIX, version=version)
    sqlite_path = prefix_directory_join(PREFIX, name="itis.sqlite", version=version)
    if not os.path.exists(sqlite_path):
        with zipfile.ZipFile(zip_path) as zip_file:
            for file in zip_file.filelist:
                if file.filename.endswith(".sqlite") and not file.is_dir():
                    zip_file.extract(file, sqlite_dir)
                    shutil.move(os.path.join(sqlite_dir, file.filename), sqlite_path)
                    os.rmdir(os.path.join(sqlite_dir, os.path.dirname(file.filename)))

    if not os.path.exists(sqlite_path):
        raise FileNotFoundError(f"file missing: {sqlite_path}")

    conn = sqlite3.connect(sqlite_path.as_posix())

    with closing(conn.cursor()) as cursor:
        cursor.execute(LONGNAMES_QUERY)
        id_to_reference = {
            str(identifier): Reference(prefix=PREFIX, identifier=str(identifier), name=name)
            for identifier, name in cursor.fetchall()
        }

    with closing(conn.cursor()) as cursor:
        cursor.execute(HIERARCHY_QUERY)
        id_to_parents = multidict((str(child), str(parent)) for child, parent in cursor.fetchall())

    for identifier, reference in id_to_reference.items():
        parents = []
        for parent_identifier in id_to_parents.get(identifier, []):
            if parent_identifier == "0":  # this means it's a plant
                continue
            parents.append(id_to_reference[parent_identifier])
        term = Term(
            reference=reference,
            parents=parents,
        )
        yield term


if __name__ == "__main__":
    ITISGetter.cli()



# -*- coding: utf-8 -*-

"""CLI for PyOBO lookups."""

import json
import sys
from typing import Optional

import bioregistry
import click
from more_click import verbose_option

from .utils import (
    echo_df,
    force_option,
    no_strict_option,
    prefix_argument,
    version_option,
)
from ..api import (
    get_ancestors,
    get_descendants,
    get_filtered_properties_df,
    get_filtered_relations_df,
    get_filtered_xrefs,
    get_hierarchy,
    get_id_definition_mapping,
    get_id_name_mapping,
    get_id_species_mapping,
    get_id_synonyms_mapping,
    get_id_to_alts,
    get_ids,
    get_metadata,
    get_name,
    get_name_by_curie,
    get_properties_df,
    get_relations_df,
    get_typedef_df,
    get_xrefs_df,
)
from ..identifier_utils import normalize_curie

__all__ = [
    "lookup",
]


@click.group()
def lookup():
    """Lookup resources."""


@lookup.command()
@prefix_argument
@click.option("-t", "--target")
@verbose_option
@force_option
@no_strict_option
@version_option
def xrefs(prefix: str, target: str, force: bool, no_strict: bool, version: Optional[str]):
    """Page through xrefs for the given namespace to the second given namespace."""
    if target:
        target = bioregistry.normalize_prefix(target)
        filtered_xrefs = get_filtered_xrefs(
            prefix, target, force=force, strict=not no_strict, version=version
        )
        click.echo_via_pager(
            "\n".join(f"{identifier}\t{_xref}" for identifier, _xref in filtered_xrefs.items())
        )
    else:
        all_xrefs_df = get_xrefs_df(prefix, force=force, strict=not no_strict, version=version)
        echo_df(all_xrefs_df)


@lookup.command()
@prefix_argument
@verbose_option
@force_option
def metadata(prefix: str, force: bool):
    """Print the metadata for the given namespace."""
    metadata = get_metadata(prefix, force=force)
    click.echo(json.dumps(metadata, indent=2))


@lookup.command()
@prefix_argument
@verbose_option
@force_option
@no_strict_option
@version_option
def ids(prefix: str, force: bool, no_strict: bool, version: Optional[str]):
    """Page through the identifiers of entities in the given namespace."""
    id_list = get_ids(prefix, force=force, strict=not no_strict, version=version)
    click.echo_via_pager("\n".join(id_list))


@lookup.command()
@prefix_argument
@verbose_option
@force_option
@no_strict_option
@click.option("-i", "--identifier")
@version_option
def names(
    prefix: str, identifier: Optional[str], force: bool, no_strict: bool, version: Optional[str]
):
    """Page through the identifiers and names of entities in the given namespace."""
    id_to_name = get_id_name_mapping(prefix, force=force, strict=not no_strict, version=version)
    if identifier is None:
        _help_page_mapping(id_to_name)
    else:
        name = id_to_name.get(identifier)
        if name is None:
            click.secho(f"No name available for {identifier}", fg="red")
        else:
            click.echo(name)


@lookup.command()
@prefix_argument
@verbose_option
@force_option
@no_strict_option
@click.option("-i", "--identifier")
@version_option
def species(
    prefix: str, identifier: Optional[str], force: bool, no_strict: bool, version: Optional[str]
):
    """Page through the identifiers and species of entities in the given namespace."""
    id_to_species = get_id_species_mapping(
        prefix, force=force, strict=not no_strict, version=version
    )
    if identifier is None:
        _help_page_mapping(id_to_species)
    else:
        species = id_to_species.get(identifier)
        if species is None:
            click.secho(f"No species available for {identifier}", fg="red")
        else:
            click.echo(species)


@lookup.command()
@prefix_argument
@verbose_option
@force_option
@click.option("-i", "--identifier")
@version_option
def definitions(prefix: str, identifier: Optional[str], force: bool, version: Optional[str]):
    """Page through the identifiers and definitions of entities in the given namespace."""
    id_to_definition = get_id_definition_mapping(prefix, force=force, version=version)
    if identifier is None:
        _help_page_mapping(id_to_definition)
    else:
        definition = id_to_definition.get(identifier)
        if definition is None:
            click.secho(f"No definition available for {identifier}", fg="red")
        else:
            click.echo(definition)


@lookup.command()
@prefix_argument
@verbose_option
@force_option
@version_option
def typedefs(prefix: str, force: bool, version: Optional[str]):
    """Page through the identifiers and names of typedefs in the given namespace."""
    df = get_typedef_df(prefix, force=force, version=version)
    echo_df(df)


def _help_page_mapping(id_to_name):
    click.echo_via_pager("\n".join("\t".join(item) for item in id_to_name.items()))


@lookup.command()
@prefix_argument
@verbose_option
@force_option
@version_option
def synonyms(prefix: str, force: bool, version: Optional[str]):
    """Page through the synonyms for entities in the given namespace."""
    if ":" in prefix:
        _prefix, identifier = normalize_curie(prefix)
        if _prefix is None or identifier is None:
            click.secho(f"could not normalize {prefix}")
            return sys.exit(1)
        name = get_name(_prefix, identifier)
        id_to_synonyms = get_id_synonyms_mapping(_prefix, force=force)
        click.echo(f"Synonyms for {_prefix}:{identifier} ! {name}")
        for synonym in id_to_synonyms.get(identifier, []):
            click.echo(synonym)
    else:  # it's a prefix
        id_to_synonyms = get_id_synonyms_mapping(prefix, force=force, version=version)
        click.echo_via_pager(
            "\n".join(
                f"{identifier}\t{_synonym}"
                for identifier, _synonyms in id_to_synonyms.items()
                for _synonym in _synonyms
            )
        )


@lookup.command()
@prefix_argument
@click.option(
    "--relation", help="CURIE for the relationship or just the ID if local to the ontology"
)
@click.option("--target", help="Prefix for the target")
@verbose_option
@no_strict_option
@force_option
@click.option("--summarize", is_flag=True)
@version_option
def relations(
    prefix: str,
    relation: str,
    target: str,
    force: bool,
    no_strict: bool,
    summarize: bool,
    version: Optional[str],
):
    """Page through the relations for entities in the given namespace."""
    if relation is None:
        relations_df = get_relations_df(prefix, force=force, strict=not no_strict, version=version)
        if summarize:
            click.echo(relations_df[relations_df.columns[2]].value_counts())
        else:
            echo_df(relations_df)
    else:
        curie = normalize_curie(relation)
        if curie[1] is None:  # that's the identifier
            click.secho(f"not valid curie, assuming local to {prefix}", fg="yellow")
            curie = prefix, relation

        if target is not None:
            norm_target = bioregistry.normalize_prefix(target)
            if norm_target is None:
                raise ValueError
            relations_df = get_filtered_relations_df(
                prefix, relation=curie, force=force, strict=not no_strict, target=norm_target
            )
        else:
            raise NotImplementedError(f"can not filter by target prefix {target}")


@lookup.command()
@prefix_argument
@click.option("--include-part-of", is_flag=True)
@click.option("--include-has-member", is_flag=True)
@verbose_option
@force_option
@version_option
def hierarchy(
    prefix: str,
    include_part_of: bool,
    include_has_member: bool,
    force: bool,
    version: Optional[str],
):
    """Page through the hierarchy for entities in the namespace."""
    h = get_hierarchy(
        prefix,
        include_part_of=include_part_of,
        include_has_member=include_has_member,
        force=force,
        version=version,
    )
    click.echo_via_pager("\n".join("\t".join(row) for row in h.edges()))


@lookup.command()
@prefix_argument
@click.argument("identifier")
@verbose_option
@force_option
@version_option
def ancestors(prefix: str, identifier: str, force: bool, version: Optional[str]):
    """Look up ancestors."""
    curies = get_ancestors(prefix=prefix, identifier=identifier, force=force, version=version)
    for curie in sorted(curies or []):
        click.echo(f"{curie}\t{get_name_by_curie(curie)}")


@lookup.command()
@prefix_argument
@click.argument("identifier")
@verbose_option
@force_option
@version_option
def descendants(prefix: str, identifier: str, force: bool, version: Optional[str]):
    """Look up descendants."""
    curies = get_descendants(prefix=prefix, identifier=identifier, force=force, version=version)
    for curie in sorted(curies or []):
        click.echo(f"{curie}\t{get_name_by_curie(curie)}")


@lookup.command()
@prefix_argument
@click.option("-k", "--key")
@verbose_option
@force_option
@version_option
def properties(prefix: str, key: Optional[str], force: bool, version: Optional[str]):
    """Page through the properties for entities in the given namespace."""
    if key is None:
        properties_df = get_properties_df(prefix, force=force, version=version)
    else:
        properties_df = get_filtered_properties_df(prefix, prop=key, force=force)
    echo_df(properties_df)


@lookup.command()
@prefix_argument
@verbose_option
@force_option
@click.option("-i", "--identifier")
@version_option
def alts(prefix: str, identifier: Optional[str], force: bool, version: Optional[str]):
    """Page through alt ids in a namespace."""
    id_to_alts = get_id_to_alts(prefix, force=force, version=version)
    if identifier is None:
        click.echo_via_pager(
            "\n".join(
                f"{identifier}\t{alt}" for identifier, alts in id_to_alts.items() for alt in alts
            )
        )
    else:
        _alts = id_to_alts.get(identifier)
        if _alts is None:
            click.secho(f"No alternate identifiers for {identifier}", fg="red")
        else:
            click.echo("\n".join(_alts))



# -*- coding: utf-8 -*-

"""Tests for PyOBO grounding."""

import unittest

import pyobo
from pyobo.mocks import get_mock_id_name_mapping, get_mock_id_synonyms_mapping

mock_id_name_mapping = get_mock_id_name_mapping(
    {
        "chebi": {
            "132964": "fluazifop-P-butyl",
        },
    }
)

mock_id_synonyms_mapping = get_mock_id_synonyms_mapping(
    {
        "chebi": {
            "132964": ["Fusilade II"],
        },
    }
)


class TestGround(unittest.TestCase):
    """Test grounding."""

    def test_ground(self):
        """Test grounding a ChEBI entry by name and synonym."""
        for query in ("Fusilade II", "fluazifop-P-butyl"):
            with self.subTest(query=query), mock_id_name_mapping, mock_id_synonyms_mapping:
                prefix, identifier, name = pyobo.ground("chebi", query)
                self.assertIsNotNone(prefix)
                self.assertIsNotNone(identifier)
                self.assertIsNotNone(name)
                self.assertEqual("chebi", prefix)
                self.assertEqual("132964", identifier)
                self.assertEqual("fluazifop-P-butyl", name)



# -*- coding: utf-8 -*-

"""Utilities for caching files."""

import gzip
import json
import logging
import os
from pathlib import Path
from typing import Generic, Iterable, List, Mapping, TypeVar, Union

import networkx as nx
from pystow.cache import Cached
from pystow.cache import CachedCollection as cached_collection  # noqa:N813
from pystow.cache import CachedDataFrame as cached_df  # noqa:N813
from pystow.cache import CachedJSON as cached_json  # noqa:N813
from pystow.cache import CachedPickle as cached_pickle  # noqa:N813

from .io import open_map_tsv, open_multimap_tsv, write_map_tsv, write_multimap_tsv

__all__ = [
    # from pystow
    "cached_json",
    "cached_collection",
    "cached_df",
    "cached_pickle",
    # implemented here
    "cached_graph",
    "cached_mapping",
    "cached_multidict",
]

logger = logging.getLogger(__name__)

X = TypeVar("X")


class _CachedMapping(Cached[X], Generic[X]):
    """A cache for simple mappings."""

    def __init__(
        self,
        path: Union[str, Path, os.PathLike],
        header: Iterable[str],
        *,
        use_tqdm: bool = False,
        force: bool = False,
    ):
        super().__init__(path=path, force=force)
        self.header = header
        self.use_tqdm = use_tqdm


class CachedMapping(_CachedMapping[Mapping[str, str]]):
    """A cache for simple mappings."""

    def load(self) -> Mapping[str, str]:
        return open_map_tsv(self.path, use_tqdm=self.use_tqdm)

    def dump(self, rv: Mapping[str, str]) -> None:
        write_map_tsv(path=self.path, header=self.header, rv=rv)


cached_mapping = CachedMapping


def get_gzipped_graph(path: Union[str, Path]) -> nx.MultiDiGraph:
    """Read a graph that's gzipped nodelink."""
    with gzip.open(path, "rt") as file:
        return nx.node_link_graph(json.load(file))


def write_gzipped_graph(graph: nx.MultiDiGraph, path: Union[str, Path]) -> None:
    """Write a graph as gzipped nodelink."""
    with gzip.open(path, "wt") as file:
        json.dump(nx.node_link_data(graph), file)


class CachedGraph(Cached[nx.MultiDiGraph]):
    def load(self) -> nx.MultiDiGraph:
        return get_gzipped_graph(self.path)

    def dump(self, rv: nx.MultiDiGraph) -> None:
        write_gzipped_graph(rv, self.path)


cached_graph = CachedGraph


class CachedMultidict(_CachedMapping[Mapping[str, List[str]]]):
    """A cache for complex mappings."""

    def load(self) -> Mapping[str, List[str]]:
        return open_multimap_tsv(self.path, use_tqdm=self.use_tqdm)

    def dump(self, rv: Mapping[str, List[str]]) -> None:
        write_multimap_tsv(path=self.path, header=self.header, rv=rv)


cached_multidict = CachedMultidict



# -*- coding: utf-8 -*-

"""Convert KEGG Pathways to OBO.

Run with ``python -m pyobo.sources.kegg.pathway``
"""

import logging
import urllib.error
from collections import defaultdict
from functools import partial
from typing import Iterable, List, Mapping, Tuple, Union

from tqdm.auto import tqdm
from tqdm.contrib.concurrent import thread_map
from tqdm.contrib.logging import logging_redirect_tqdm

from pyobo.sources.kegg.api import (
    KEGG_GENES_PREFIX,
    KEGG_PATHWAY_PREFIX,
    KEGGGenome,
    ensure_link_pathway_genome,
    ensure_list_pathway_genome,
    ensure_list_pathways,
)
from pyobo.sources.kegg.genome import iter_kegg_genomes
from pyobo.struct import (
    Obo,
    Reference,
    Term,
    from_species,
    has_participant,
    species_specific,
)

__all__ = [
    "KEGGPathwayGetter",
]

logger = logging.getLogger(__name__)


class KEGGPathwayGetter(Obo):
    """An ontology representation of KEGG Pathways."""

    ontology = KEGG_PATHWAY_PREFIX
    bioversions_key = "kegg"
    typedefs = [from_species, species_specific, has_participant]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise)


def get_obo() -> Obo:
    """Get KEGG Pathways as OBO."""
    # since old kegg versions go away forever, do NOT add a force option
    return KEGGPathwayGetter()


def iter_terms(version: str, skip_missing: bool = True) -> Iterable[Term]:
    """Iterate over terms for KEGG Pathway."""
    # since old kegg versions go away forever, do NOT add a force option
    yield from _iter_map_terms(version=version)
    it = iter_kegg_pathway_paths(version=version, skip_missing=skip_missing)
    for row in tqdm(it, unit_scale=True, unit="genome", desc="Parsing genomes"):
        if not row:
            continue
        kegg_genome, list_pathway_path, link_pathway_path = row
        if not kegg_genome or not list_pathway_path or not link_pathway_path:
            continue
        yield from _iter_genome_terms(
            list_pathway_path=list_pathway_path,
            link_pathway_path=link_pathway_path,
            kegg_genome=kegg_genome,
        )


def _get_link_pathway_map(path: str) -> Mapping[str, List[str]]:
    rv = defaultdict(list)
    with open(path) as file:
        for line in file:
            try:
                protein_id, pathway_id = line.strip().split("\t")
            except ValueError:
                logger.warning("Unable to parse link file line: %s", line)
            else:
                rv[pathway_id[len("path:") :]].append(protein_id)

    return {pathway_id: sorted(protein_ids) for pathway_id, protein_ids in rv.items()}


def _iter_map_terms(version: str) -> Iterable[Term]:
    for identifier, name in ensure_list_pathways(version=version).items():
        yield Term.from_triple(
            prefix=KEGG_PATHWAY_PREFIX,
            identifier=identifier,
            name=name,
        )


def _iter_genome_terms(
    *,
    list_pathway_path: str,
    link_pathway_path: str,
    kegg_genome: KEGGGenome,
) -> Iterable[Term]:
    terms = {}
    with open(list_pathway_path) as file:
        list_pathway_lines = [line.strip() for line in file]
    for line in list_pathway_lines:
        line = line.strip()
        pathway_id, name = [part.strip() for part in line.split("\t")]
        pathway_id = pathway_id[len("path:") :]

        terms[pathway_id] = term = Term.from_triple(
            prefix=KEGG_PATHWAY_PREFIX,
            identifier=pathway_id,
            name=name,
        )

        # Annotate species information
        kegg_genome.annotate_term(term)

        # Annotate the non-species specific code
        _start = min(i for i, e in enumerate(pathway_id) if e.isnumeric())
        pathway_code = pathway_id[_start:]
        term.append_relationship(
            species_specific,
            Reference(prefix=KEGG_PATHWAY_PREFIX, identifier=f"map{pathway_code}"),
        )

    for pathway_id, protein_ids in _get_link_pathway_map(link_pathway_path).items():
        pathway_term = terms.get(pathway_id)
        if pathway_term is None:
            tqdm.write(f"could not find kegg.pathway:{pathway_id} for {kegg_genome.name}")
            continue
        for protein_id in protein_ids:
            pathway_term.append_relationship(
                has_participant,
                Reference(
                    prefix=KEGG_GENES_PREFIX,
                    identifier=protein_id,
                ),
            )

    yield from terms.values()


def iter_kegg_pathway_paths(
    version: str, skip_missing: bool = True
) -> Iterable[Union[Tuple[KEGGGenome, str, str], Tuple[None, None, None]]]:
    """Get paths for the KEGG Pathway files."""
    genomes = list(iter_kegg_genomes(version=version, desc="KEGG Pathways"))
    func = partial(_process_genome, version=version, skip_missing=skip_missing)
    return thread_map(func, genomes, unit="pathway", unit_scale=True)


def _process_genome(kegg_genome, version, skip_missing):
    with logging_redirect_tqdm():
        try:
            list_pathway_path = ensure_list_pathway_genome(
                kegg_genome.identifier,
                version=version,
                error_on_missing=not skip_missing,
            )
            link_pathway_path = ensure_link_pathway_genome(
                kegg_genome.identifier,
                version=version,
                error_on_missing=not skip_missing,
            )
        except urllib.error.HTTPError as e:
            code = e.getcode()
            if code != 404:
                msg = (
                    f"[HTTP {code}] Error downloading {kegg_genome.identifier} ({kegg_genome.name}"
                )
                if kegg_genome.taxonomy_id is None:
                    msg = f"{msg}): {e.geturl()}"
                else:
                    msg = f"{msg}; taxonomy:{kegg_genome.taxonomy_id}): {e.geturl()}"
                tqdm.write(msg)
        except FileNotFoundError:
            return None, None, None
        else:
            return kegg_genome, list_pathway_path, link_pathway_path


if __name__ == "__main__":
    KEGGPathwayGetter.cli()



# -*- coding: utf-8 -*-

"""CLI for PyOBO Database Generation."""

import logging
from typing import Optional

import click
from more_click import verbose_option
from tqdm.contrib.logging import logging_redirect_tqdm
from zenodo_client import update_zenodo

from .utils import directory_option, force_option, no_strict_option, zenodo_option
from ..constants import (
    ALTS_DATA_RECORD,
    DEFINITIONS_RECORD,
    JAVERT_RECORD,
    OOH_NA_NA_RECORD,
    PROPERTIES_RECORD,
    RELATIONS_RECORD,
    SPECIES_RECORD,
    SYNONYMS_RECORD,
    TYPEDEFS_RECORD,
)
from ..getters import db_output_helper
from ..xrefdb.xrefs_pipeline import (
    _iter_alts,
    _iter_definitions,
    _iter_metadata,
    _iter_names,
    _iter_properties,
    _iter_relations,
    _iter_species,
    _iter_synonyms,
    _iter_typedefs,
    _iter_xrefs,
)

__all__ = [
    "main",
]


@click.group(name="database")
def main():
    """Build the PyOBO Database."""


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
@click.pass_context
def build(ctx: click.Context, directory: str, zenodo: bool, no_strict: bool, force: bool):
    """Build all databases."""
    # if no_strict and zenodo:
    #    click.secho("Must be strict before uploading", fg="red")
    #    sys.exit(1)
    with logging_redirect_tqdm():
        click.secho("Collecting metadata and building", fg="cyan", bold=True)
        # note that this is the only one that needs a force=force
        ctx.invoke(metadata, directory=directory, no_strict=no_strict, force=force)
        click.secho("Alternate Identifiers", fg="cyan", bold=True)
        ctx.invoke(alts, directory=directory, zenodo=zenodo, no_strict=no_strict)
        click.secho("Synonyms", fg="cyan", bold=True)
        ctx.invoke(synonyms, directory=directory, zenodo=zenodo, no_strict=no_strict)
        click.secho("Xrefs", fg="cyan", bold=True)
        ctx.invoke(xrefs, directory=directory, zenodo=zenodo, no_strict=no_strict)
        click.secho("Names", fg="cyan", bold=True)
        ctx.invoke(names, directory=directory, zenodo=zenodo, no_strict=no_strict)
        click.secho("Definitions", fg="cyan", bold=True)
        ctx.invoke(definitions, directory=directory, zenodo=zenodo, no_strict=no_strict)
        click.secho("Properties", fg="cyan", bold=True)
        ctx.invoke(properties, directory=directory, zenodo=zenodo, no_strict=no_strict)
        click.secho("Relations", fg="cyan", bold=True)
        ctx.invoke(relations, directory=directory, zenodo=zenodo, no_strict=no_strict)
        click.secho("Typedefs", fg="cyan", bold=True)
        ctx.invoke(typedefs, directory=directory, zenodo=zenodo, no_strict=no_strict)
        click.secho("Species", fg="cyan", bold=True)
        ctx.invoke(species, directory=directory, zenodo=zenodo, no_strict=no_strict)


skip_below_option = click.option("--skip-below")
skip_below_exclusive_option = click.option("--skip-below-exclusive", is_flag=True)


@main.command()
@verbose_option
@directory_option
@force_option
@no_strict_option
@skip_below_option
@click.option("--skip-pyobo")
def metadata(
    directory: str, no_strict: bool, force: bool, skip_below: Optional[str], skip_pyobo: bool
):
    """Make the prefix-metadata dump."""
    db_output_helper(
        _iter_metadata,
        "metadata",
        ("prefix", "version", "date", "deprecated"),
        strict=not no_strict,
        force=force,
        directory=directory,
        use_gzip=False,
        skip_below=skip_below,
        skip_pyobo=skip_pyobo,
    )


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
@skip_below_option
@skip_below_exclusive_option
def names(
    directory: str,
    zenodo: bool,
    no_strict: bool,
    force: bool,
    skip_below: Optional[str],
    skip_below_exclusive: bool,
):
    """Make the prefix-identifier-name dump."""
    with logging_redirect_tqdm():
        paths = db_output_helper(
            _iter_names,
            "names",
            ("prefix", "identifier", "name"),
            strict=not no_strict,
            force=force,
            directory=directory,
            skip_below=skip_below,
            skip_below_inclusive=not skip_below_exclusive,
        )
    if zenodo:
        # see https://zenodo.org/record/4020486
        update_zenodo(OOH_NA_NA_RECORD, paths)


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
def species(directory: str, zenodo: bool, no_strict: bool, force: bool):
    """Make the prefix-identifier-species dump."""
    with logging_redirect_tqdm():
        paths = db_output_helper(
            _iter_species,
            "species",
            ("prefix", "identifier", "species"),
            strict=not no_strict,
            force=force,
            directory=directory,
        )
    if zenodo:
        # see https://zenodo.org/record/5334738
        update_zenodo(SPECIES_RECORD, paths)


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
def definitions(directory: str, zenodo: bool, no_strict: bool, force: bool):
    """Make the prefix-identifier-definition dump."""
    with logging_redirect_tqdm():
        paths = db_output_helper(
            _iter_definitions,
            "definitions",
            ("prefix", "identifier", "definition"),
            strict=not no_strict,
            force=force,
            directory=directory,
            skip_set={"kegg.pathway", "kegg.genes", "kegg.genome", "umls"},
        )
    if zenodo:
        # see https://zenodo.org/record/4637061
        update_zenodo(DEFINITIONS_RECORD, paths)


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
def typedefs(directory: str, zenodo: bool, no_strict: bool, force: bool):
    """Make the typedef prefix-identifier-name dump."""
    with logging_redirect_tqdm():
        paths = db_output_helper(
            _iter_typedefs,
            "typedefs",
            ("prefix", "typedef_prefix", "identifier", "name"),
            strict=not no_strict,
            force=force,
            directory=directory,
            use_gzip=False,
            skip_set={"ncbigene", "kegg.pathway", "kegg.genes", "kegg.genome"},
        )
    if zenodo:
        # see https://zenodo.org/record/4644013
        update_zenodo(TYPEDEFS_RECORD, paths)


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
def alts(directory: str, zenodo: bool, force: bool, no_strict: bool):
    """Make the prefix-alt-id dump."""
    with logging_redirect_tqdm():
        paths = db_output_helper(
            _iter_alts,
            "alts",
            ("prefix", "identifier", "alt"),
            directory=directory,
            force=force,
            strict=not no_strict,
            skip_set={"kegg.pathway", "kegg.genes", "kegg.genome", "umls"},
        )
    if zenodo:
        # see https://zenodo.org/record/4021476
        update_zenodo(ALTS_DATA_RECORD, paths)


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
def synonyms(directory: str, zenodo: bool, force: bool, no_strict: bool):
    """Make the prefix-identifier-synonym dump."""
    with logging_redirect_tqdm():
        paths = db_output_helper(
            _iter_synonyms,
            "synonyms",
            ("prefix", "identifier", "synonym"),
            directory=directory,
            force=force,
            strict=not no_strict,
            skip_set={"kegg.pathway", "kegg.genes", "kegg.genome"},
        )
    if zenodo:
        # see https://zenodo.org/record/4021482
        update_zenodo(SYNONYMS_RECORD, paths)


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
def relations(directory: str, zenodo: bool, force: bool, no_strict: bool):
    """Make the relation dump."""
    with logging_redirect_tqdm():
        paths = db_output_helper(
            _iter_relations,
            "relations",
            (
                "source_prefix",
                "source_identifier",
                "relation_prefix",
                "relation_identifier",
                "target_prefix",
                "target_identifier",
            ),
            directory=directory,
            force=force,
            strict=not no_strict,
            summary_detailed=(0, 2, 3),  # second column corresponds to relation type
        )
    if zenodo:
        # see https://zenodo.org/record/4625167
        update_zenodo(RELATIONS_RECORD, paths)


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
def properties(directory: str, zenodo: bool, force: bool, no_strict: bool):
    """Make the properties dump."""
    with logging_redirect_tqdm():
        paths = db_output_helper(
            _iter_properties,
            "properties",
            ("prefix", "identifier", "property", "value"),
            directory=directory,
            force=force,
            strict=not no_strict,
            summary_detailed=(0, 2),  # second column corresponds to property type
        )
    if zenodo:
        # see https://zenodo.org/record/4625172
        update_zenodo(PROPERTIES_RECORD, paths)


@main.command()
@verbose_option
@directory_option
@zenodo_option
@force_option
@no_strict_option
def xrefs(directory: str, zenodo: bool, force: bool, no_strict: bool):  # noqa: D202
    """Make the prefix-identifier-xref dump."""
    with logging_redirect_tqdm():
        paths = db_output_helper(
            _iter_xrefs,
            "xrefs",
            ("prefix", "identifier", "xref_prefix", "xref_identifier", "provenance"),
            directory=directory,
            force=force,
            strict=not no_strict,
            summary_detailed=(0, 2),  # second column corresponds to xref prefix
        )
    if zenodo:
        # see https://zenodo.org/record/4021477
        update_zenodo(JAVERT_RECORD, paths)


if __name__ == "__main__":
    logging.captureWarnings(True)
    with logging_redirect_tqdm():
        main()



# -*- coding: utf-8 -*-

"""CLI for PyOBO's Gilda Service."""

import click
from more_click import (
    host_option,
    port_option,
    run_app,
    verbose_option,
    with_gunicorn_option,
    workers_option,
)

__all__ = [
    "main",
]


@click.command(name="gilda")
@click.argument("prefix", nargs=-1)
@verbose_option
@host_option
@workers_option
@port_option
@with_gunicorn_option
def main(prefix: str, host: str, port: str, with_gunicorn: bool, workers: int):
    """Run the Gilda service for this database."""
    from .app import get_app

    app = get_app(prefix)
    run_app(app, host=host, port=port, with_gunicorn=with_gunicorn, workers=workers)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""High-level API for accessing content."""

from .alts import (  # noqa: F401
    get_alts_to_id,
    get_id_to_alts,
    get_primary_curie,
    get_primary_identifier,
)
from .hierarchy import (  # noqa: F401
    get_ancestors,
    get_descendants,
    get_hierarchy,
    get_subhierarchy,
    has_ancestor,
    is_descendent,
)
from .metadata import get_metadata  # noqa: F401
from .names import (  # noqa: F401
    get_definition,
    get_id_definition_mapping,
    get_id_name_mapping,
    get_id_synonyms_mapping,
    get_ids,
    get_name,
    get_name_by_curie,
    get_name_id_mapping,
    get_synonyms,
)
from .properties import (  # noqa: F401
    get_filtered_properties_df,
    get_filtered_properties_mapping,
    get_filtered_properties_multimapping,
    get_properties,
    get_properties_df,
    get_property,
)
from .relations import (  # noqa: F401
    get_filtered_relations_df,
    get_id_multirelations_mapping,
    get_relation,
    get_relation_mapping,
    get_relations_df,
)
from .species import get_id_species_mapping, get_species  # noqa: F401
from .typedefs import get_typedef_df  # noqa: F401
from .xrefs import (  # noqa: F401
    get_filtered_xrefs,
    get_sssom_df,
    get_xref,
    get_xrefs,
    get_xrefs_df,
)



# -*- coding: utf-8 -*-

"""Converter for GWAS Central Phenotypes."""

import json
from typing import Iterable

from tqdm.auto import tqdm, trange

from pyobo.sources.gwascentral_study import VERSION
from pyobo.struct import Obo, Reference, Term
from pyobo.utils.path import ensure_path

__all__ = [
    "GWASCentralPhenotypeGetter",
]

PREFIX = "gwascentral.phenotype"


class GWASCentralPhenotypeGetter(Obo):
    """An ontology representation of GWAS Central's phenotype nomenclature."""

    ontology = PREFIX
    static_version = VERSION

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get GWAS Central Studies as OBO."""
    return GWASCentralPhenotypeGetter(force=force)


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate over terms from GWAS Central Phenotype."""
    for n in trange(1, 11000, desc=f"{PREFIX} download"):
        try:
            path = ensure_path(
                PREFIX,
                "phenotype",
                version=version,
                url=f"https://www.gwascentral.org/phenotype/HGVPM{n}?format=json",
                name=f"HGVPM{n}.json",
                force=force,
            )
        except OSError as e:
            tqdm.write(f"{n}: {e}")
            continue
        with open(path) as file:
            j = json.load(file)

        description = j.get("description")
        if description is not None:
            description = description.strip().replace("\n", " ")
        term = Term(
            reference=Reference(prefix=PREFIX, identifier=j["identifier"], name=j["name"]),
            definition=description,
        )
        yield term


if __name__ == "__main__":
    GWASCentralPhenotypeGetter.cli()



# -*- coding: utf-8 -*-

"""Convert KEGG Genome to OBO.

Run with ``python -m pyobo.sources.kegg.genome``
"""

import logging
from typing import Iterable

from tqdm.auto import tqdm

from pyobo.constants import NCBITAXON_PREFIX
from pyobo.resources.ncbitaxon import get_ncbitaxon_name
from pyobo.sources.kegg.api import (
    KEGG_GENOME_PREFIX,
    SKIP,
    KEGGGenome,
    ensure_list_genomes,
)
from pyobo.struct import Obo, Reference, Term

__all__ = [
    "KEGGGenomeGetter",
]

logger = logging.getLogger(__name__)


def _s(line: str, sep: str):
    return [part.strip() for part in line.strip().split(sep, 1)]


class KEGGGenomeGetter(Obo):
    """An ontology representation of KEGG Genomes."""

    ontology = KEGG_GENOME_PREFIX
    bioversions_key = "kegg"

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise)


def get_obo() -> Obo:
    """Get KEGG Genome as OBO."""
    # since old kegg versions go away forever, do NOT add a force option
    return KEGGGenomeGetter()


def parse_genome_line(line: str) -> KEGGGenome:
    """Parse a line from the KEGG Genome database."""
    line = line.strip()
    identifier, rest = _s(line, "\t")
    identifier = identifier[len("gn:") :]
    if ";" in rest:
        rest, name = _s(rest, ";")

        rest = [part.strip() for part in rest.split(",")]
        if len(rest) == 3:
            kegg_code, long_code, taxonomy_id = rest
        elif len(rest) == 2:
            kegg_code, taxonomy_id = rest
            long_code = None
        elif len(rest) == 1:
            (kegg_code,) = rest
            long_code, taxonomy_id = None, None
        else:
            raise ValueError(f"unexpected line: {line}")
    else:
        name = rest
        taxonomy_id = None
        long_code = None
        kegg_code = None

    if "\t" in name:
        # TODO maybe throw this out?
        tqdm.write(f"[{KEGG_GENOME_PREFIX}] tab in name: {name}")
        name = name.replace("\t", " ")

    return KEGGGenome(
        identifier=identifier,
        code=kegg_code,
        long_code=long_code,
        taxonomy_id=taxonomy_id,
        name=name,
    )


def iter_kegg_genomes(version: str, desc: str) -> Iterable[KEGGGenome]:
    """Iterate over all KEGG genomes."""
    # since old kegg versions go away forever, do NOT add a force option
    path = ensure_list_genomes(version=version)
    with open(path) as file:
        lines = [line.strip() for line in file]
    it = tqdm(lines, desc=desc, unit_scale=True, unit="genome")
    for line in it:
        yv = parse_genome_line(line)
        it.set_postfix({"id": yv.identifier, "name": yv.name})
        yield yv


def iter_terms(version: str) -> Iterable[Term]:
    """Iterate over terms for KEGG Genome."""
    # since old kegg versions go away forever, do NOT add a force option
    errors = 0
    for kegg_genome in iter_kegg_genomes(version=version, desc="KEGG Genomes"):
        if kegg_genome.identifier in SKIP:
            continue
        term = Term.from_triple(
            prefix=KEGG_GENOME_PREFIX,
            identifier=kegg_genome.identifier,
            name=kegg_genome.name,
        )
        if kegg_genome.taxonomy_id is not None:
            taxonomy_name = get_ncbitaxon_name(kegg_genome.taxonomy_id)
            if taxonomy_name is None:
                errors += 1
                logger.debug(
                    f"[{KEGG_GENOME_PREFIX}] could not find name for taxonomy:{kegg_genome.taxonomy_id}"
                )
            term.append_xref(
                Reference(
                    prefix=NCBITAXON_PREFIX,
                    identifier=kegg_genome.taxonomy_id,
                    name=taxonomy_name,
                )
            )
        yield term

    if errors:
        logger.info("[%s] unable to find %d taxonomy names in NCBI", KEGG_GENOME_PREFIX, errors)


if __name__ == "__main__":
    KEGGGenomeGetter.cli()



# -*- coding: utf-8 -*-

"""CLI for PyOBO's interface to S3."""

import click
from more_click import verbose_option

from ..aws import download_artifacts, list_artifacts, upload_artifacts

__all__ = [
    "main",
]

bucket_argument = click.argument("bucket")


@click.group(name="aws")
def main():
    """S3 utilities."""


@main.command()
@bucket_argument
@verbose_option
def download(bucket):
    """Download all artifacts from the S3 bucket."""
    download_artifacts(bucket)


@main.command()
@bucket_argument
@verbose_option
@click.option("-w", "--whitelist", multiple=True)
@click.option("-b", "--blacklist", multiple=True)
def upload(bucket, whitelist, blacklist):
    """Download all artifacts from the S3 bucket."""
    upload_artifacts(bucket, whitelist=whitelist, blacklist=blacklist)


@main.command()
@bucket_argument
@verbose_option
def ls(bucket):
    """List all artifacts on the S3 bucket."""
    click.echo(list_artifacts(bucket))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Converter for HGNC."""

import itertools as itt
import json
import logging
import typing
from collections import Counter, defaultdict
from operator import attrgetter
from typing import DefaultDict, Dict, Iterable, Optional

import bioversions
from tabulate import tabulate
from tqdm.auto import tqdm

from pyobo.struct import (
    Obo,
    Reference,
    Synonym,
    SynonymTypeDef,
    Term,
    from_species,
    gene_product_member_of,
    has_gene_product,
    member_of,
    orthologous,
    transcribes_to,
)
from pyobo.utils.path import ensure_path, prefix_directory_join

__all__ = [
    "HGNCGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "hgnc"
DEFINITIONS_URL_FMT = (
    "https://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/"
    "archive/monthly/json/hgnc_complete_set_{version}.json"
)

previous_symbol_type = SynonymTypeDef.from_text("previous_symbol")
alias_symbol_type = SynonymTypeDef.from_text("alias_symbol")
previous_name_type = SynonymTypeDef.from_text("previous_name")
alias_name_type = SynonymTypeDef.from_text("alias_name")

#: First column is MIRIAM prefix, second column is HGNC key
gene_xrefs = [
    ("ensembl", "ensembl_gene_id"),
    ("ncbigene", "entrez_id"),
    ("cosmic", "cosmic"),
    ("vega", "vega_id"),
    ("ucsc", "ucsc_id"),
    ("merops.entry", "merops"),
    ("lncipedia", "lncipedia"),
    ("orphanet", "orphanet"),
    ("pseudogene", "pseudogene.org"),
    ("ena", "ena"),
    ("refseq", "refseq_accession"),
    ("iuphar.receptor", "iuphar"),
    # ("mgi", "mgd_id"),
    ("ccds", "ccds_id"),
    # ("rgd", "rgd_id"),
    ("omim", "omim_id"),
    # ('uniprot', 'uniprot_ids'),
    # ('ec-code', 'enzyme_id'),
    # ('rnacentral', 'rna_central_id'),
    # ('mirbase', 'mirbase'),
    # ('snornabase', 'snornabase'),
]

#: Encodings from https://www.genenames.org/cgi-bin/statistics
#: To see all, do: ``cat hgnc_complete_set.json | jq .response.docs[].locus_type | sort | uniq``
ENCODINGS = {
    # protein-coding gene
    "gene with protein product": "GRP",
    # non-coding RNA
    "RNA, Y": "GR",
    "RNA, cluster": "GR",
    "RNA, long non-coding": "GR",
    "RNA, micro": "GM",
    "RNA, misc": "GR",
    "RNA, ribosomal": "GR",
    "RNA, small cytoplasmic": "GR",
    "RNA, small nuclear": "GR",
    "RNA, small nucleolar": "GR",
    "RNA, transfer": "GR",
    "RNA, vault": "GR",
    # phenotype
    "phenotype only": "G",
    # pseudogene
    "T cell receptor pseudogene": "GRP",
    "immunoglobulin pseudogene": "GRP",
    "immunoglobulin gene": "GRP",
    "pseudogene": "G",
    # other
    "T cell receptor gene": "GRP",
    "complex locus constituent": "G",
    "endogenous retrovirus": "G",
    "fragile site": "G",
    "protocadherin": "GRP",
    "readthrough": "G",
    "region": "G",
    "transposable element": "G",
    "virus integration site": "G",
    "unknown": "GRP",
}

#: A mapping from HGNC's locus_type annotations to sequence ontology identifiers
LOCUS_TYPE_TO_SO = {
    # protein-coding gene
    "gene with protein product": "0001217",
    "complex locus constituent": "0000997",  # https://github.com/pyobo/pyobo/issues/118#issuecomment-1564520052
    "protocadherin": "",  # TODO see https://github.com/The-Sequence-Ontology/SO-Ontologies/issues/562
    # non-coding RNA
    "RNA, Y": "0002359",
    "RNA, cluster": "",  # TODO see https://github.com/The-Sequence-Ontology/SO-Ontologies/issues/564
    "RNA, long non-coding": "0002127",  # HGNC links to wrong one
    "RNA, micro": "0001265",
    "RNA, misc": "0001266",
    "RNA, ribosomal": "0001637",
    "RNA, small cytoplasmic": "0001266",
    "RNA, small nuclear": "0001268",
    "RNA, small nucleolar": "0001267",
    "RNA, transfer": "0001272",
    "RNA, vault": "0002358",
    # phenotype
    "phenotype only": "0001500",  # https://github.com/pyobo/pyobo/issues/118#issuecomment-1564574892
    # pseudogene
    "T cell receptor pseudogene": "0002099",
    "immunoglobulin pseudogene": "0002098",
    "immunoglobulin gene": "0002122",  # HGNC links to wrong one
    "pseudogene": "0000336",
    # other
    "T cell receptor gene": "0002133",
    "endogenous retrovirus": "0000100",
    "fragile site": "0002349",
    "readthrough": "0000697",  # maybe not right
    "transposable element": "0000111",  # HGNC links to wrong one
    "virus integration site": "",  # TODO see https://github.com/The-Sequence-Ontology/SO-Ontologies/issues/551
    "region": "0001411",  # a small bucket for things that need a better annotation, even higher than "gene"
    "unknown": "0000704",  # gene
    None: "0000704",  # gene
}

IDSPACES = {
    prefix: f"https://bioregistry.io/{prefix}:"
    for prefix in {
        "rgd",
        "mgi",
        "eccode",
        "rnacentral",
        "pubmed",
        "uniprot",
        "mirbase",
        "snornabase",
        "hgnc",
        "hgnc.genegroup",
        "debio",
        "ensembl",
        "NCBIGene",
        "vega",
        "ucsc",
        "ena",
        "ccds",
        "omim",
        "cosmic",
        "merops",
        "orphanet",
        "pseudogene",
        "lncipedia",
        "refseq",
    }
}
IDSPACES.update(
    NCBITaxon="http://purl.obolibrary.org/obo/NCBITaxon_",
    SO="http://purl.obolibrary.org/obo/SO_",
)


class HGNCGetter(Obo):
    """An ontology representation of HGNC's gene nomenclature."""

    bioversions_key = ontology = PREFIX
    typedefs = [
        from_species,
        has_gene_product,
        gene_product_member_of,
        transcribes_to,
        orthologous,
        member_of,
    ]
    idspaces = IDSPACES
    synonym_typedefs = [
        previous_name_type,
        previous_symbol_type,
        alias_name_type,
        alias_symbol_type,
    ]
    root_terms = [
        Reference(prefix="SO", identifier=so_id) for so_id in sorted(set(LOCUS_TYPE_TO_SO.values()))
    ]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force, version=self.data_version)


def get_obo(*, force: bool = False) -> Obo:
    """Get HGNC as OBO."""
    return HGNCGetter(force=force)


def get_terms(version: Optional[str] = None, force: bool = False) -> Iterable[Term]:  # noqa:C901
    """Get HGNC terms."""
    if version is None:
        version = bioversions.get_version("hgnc")
    unhandled_entry_keys: typing.Counter[str] = Counter()
    unhandle_locus_types: DefaultDict[str, Dict[str, Term]] = defaultdict(dict)
    path = ensure_path(
        PREFIX,
        url=DEFINITIONS_URL_FMT.format(version=version),
        force=force,
        version=version,
        name="hgnc_complete_set.json",
    )
    with open(path) as file:
        entries = json.load(file)["response"]["docs"]

    yield Term.from_triple("NCBITaxon", "9606", "Homo sapiens")
    yield from sorted(
        {
            Term(reference=Reference.auto("SO", so_id))
            for so_id in sorted(LOCUS_TYPE_TO_SO.values())
            if so_id
        },
        key=attrgetter("identifier"),
    )

    statuses = set()
    for entry in tqdm(entries, desc=f"Mapping {PREFIX}", unit="gene", unit_scale=True):
        name, symbol, identifier = (
            entry.pop("name"),
            entry.pop("symbol"),
            entry.pop("hgnc_id")[len("HGNC:") :],
        )
        status = entry.pop("status")
        if status == "Approved":
            is_obsolete = False
        elif status not in statuses:
            statuses.add(status)
            tqdm.write(f"[{PREFIX}] unhandled {status}")
            is_obsolete = True
        else:
            raise ValueError(f"Unhandled status for hgnc:{identifier}: {status}")

        term = Term(
            definition=name,
            reference=Reference(prefix=PREFIX, identifier=identifier, name=symbol),
            is_obsolete=is_obsolete,
        )

        for uniprot_id in entry.pop("uniprot_ids", []):
            term.append_relationship(
                has_gene_product,
                Reference(prefix="uniprot", identifier=uniprot_id),
            )
        for ec_code in entry.pop("enzyme_id", []):
            if "-" in ec_code:
                continue  # only add concrete annotations
            term.append_relationship(
                gene_product_member_of,
                Reference(prefix="eccode", identifier=ec_code),
            )
        for rna_central_ids in entry.pop("rna_central_id", []):
            for rna_central_id in rna_central_ids.split(","):
                term.append_relationship(
                    transcribes_to,
                    Reference(prefix="rnacentral", identifier=rna_central_id.strip()),
                )
        mirbase_id = entry.pop("mirbase", None)
        if mirbase_id:
            term.append_relationship(
                transcribes_to,
                Reference(
                    prefix="mirbase",
                    identifier=mirbase_id,
                ),
            )
        snornabase_id = entry.pop("snornabase", None)
        if snornabase_id:
            term.append_relationship(
                transcribes_to, Reference(prefix="snornabase", identifier=snornabase_id)
            )

        for rgd_curie in entry.pop("rgd_id", []):
            if not rgd_curie.startswith("RGD:"):
                tqdm.write(f"hgnc:{identifier} had bad RGD CURIE: {rgd_curie}")
                continue
            rgd_id = rgd_curie[len("RGD:") :]
            term.append_relationship(
                orthologous,
                Reference(prefix="rgd", identifier=rgd_id),
            )
        for mgi_curie in entry.pop("mgd_id", []):
            if not mgi_curie.startswith("MGI:"):
                tqdm.write(f"hgnc:{identifier} had bad MGI CURIE: {mgi_curie}")
                continue
            mgi_id = mgi_curie[len("MGI:") :]
            if not mgi_id:
                continue
            term.append_relationship(
                orthologous,
                Reference(prefix="mgi", identifier=mgi_id),
            )

        iuphar = entry.pop("iuphar", None)
        if iuphar:
            if iuphar.startswith("objectId:"):
                term.append_exact_match(
                    Reference(prefix="iuphar.receptor", identifier=iuphar[len("objectId:") :])
                )
            elif iuphar.startswith("ligandId:"):
                term.append_exact_match(
                    Reference(prefix="iuphar.ligand", identifier=iuphar[len("ligandId:") :])
                )
            else:
                tqdm.write(f"unhandled IUPHAR: {iuphar}")

        for xref_prefix, key in gene_xrefs:
            xref_identifiers = entry.pop(key, None)
            if xref_identifiers is None:
                continue

            if isinstance(xref_identifiers, (str, int)):
                term.append_exact_match(
                    Reference(prefix=xref_prefix, identifier=str(xref_identifiers))
                )
            elif isinstance(xref_identifiers, list):
                if len(xref_identifiers) == 1:
                    term.append_exact_match(
                        Reference(prefix=xref_prefix, identifier=str(xref_identifiers[0]))
                    )
                else:
                    for xref_identifier in xref_identifiers:
                        term.append_xref(
                            Reference(prefix=xref_prefix, identifier=str(xref_identifier))
                        )
            else:
                raise TypeError

        for pubmed_id in entry.pop("pubmed_id", []):
            term.append_provenance(Reference(prefix="pubmed", identifier=str(pubmed_id)))

        gene_group_ids = entry.pop("gene_group_id", [])
        gene_groups = entry.pop("gene_group", [])
        for gene_group_id, gene_group_label in zip(gene_group_ids, gene_groups):
            term.append_relationship(
                member_of,
                Reference(
                    prefix="hgnc.genegroup",
                    identifier=str(gene_group_id),
                    name=gene_group_label,
                ),
            )

        for alias_symbol in entry.pop("alias_symbol", []):
            term.append_synonym(Synonym(name=alias_symbol, type=alias_symbol_type))
        for alias_name in entry.pop("alias_name", []):
            term.append_synonym(Synonym(name=alias_name, type=alias_name_type))
        for previous_symbol in itt.chain(
            entry.pop("previous_symbol", []), entry.pop("prev_symbol", [])
        ):
            term.append_synonym(Synonym(name=previous_symbol, type=previous_symbol_type))
        for previous_name in entry.pop("prev_name", []):
            term.append_synonym(Synonym(name=previous_name, type=previous_name_type))

        for prop in ["location"]:
            value = entry.pop(prop, None)
            if value:
                term.append_property(prop, value)

        locus_type = entry.pop("locus_type")
        locus_group = entry.pop("locus_group")
        so_id = LOCUS_TYPE_TO_SO.get(locus_type)
        if so_id:
            term.append_parent(Reference.auto("SO", so_id))
        else:
            term.append_parent(Reference.auto("SO", "0000704"))  # gene
            unhandle_locus_types[locus_type][identifier] = term
            term.append_property("locus_type", locus_type)
            term.append_property("locus_group", locus_group)

        term.set_species(identifier="9606", name="Homo sapiens")

        for key in entry:
            unhandled_entry_keys[key] += 1
        yield term

    with open(prefix_directory_join(PREFIX, name="unhandled.json"), "w") as file:
        json.dump(
            {
                k: {hgnc_id: term.name for hgnc_id, term in v.items()}
                for k, v in unhandle_locus_types.items()
            },
            file,
            indent=2,
        )

    with open(prefix_directory_join(PREFIX, name="unhandled.md"), "w") as file:
        for k, v in sorted(unhandle_locus_types.items()):
            t = tabulate(
                [
                    (
                        hgnc_id,
                        term.name,
                        term.is_obsolete,
                        term.bioregistry_link,
                        ", ".join(
                            p.bioregistry_link for p in term.provenance if p.bioregistry_link
                        ),
                    )
                    for hgnc_id, term in sorted(v.items())
                ],
                headers=["hgnc_id", "name", "obsolete", "link", "provenance"],
                tablefmt="github",
            )
            print(f"## {k} ({len(v)})", file=file)  # noqa: T201
            print(t, "\n", file=file)  # noqa: T201

    unhandle_locus_type_counter = Counter(
        {locus_type: len(d) for locus_type, d in unhandle_locus_types.items()}
    )
    logger.warning(
        "Unhandled locus types:\n%s", tabulate(unhandle_locus_type_counter.most_common())
    )
    logger.warning("Unhandled keys:\n%s", tabulate(unhandled_entry_keys.most_common()))


if __name__ == "__main__":
    HGNCGetter.cli()



# -*- coding: utf-8 -*-

"""Utilities or interacting with the ICD API.

Want to get your own API cliend ID and client secret?

1. Register at https://icdapihome.azurewebsites.net/icdapi/Account/Register
2. Sell your soul to the American government
"""

import datetime
import json
import os
from pathlib import Path
from typing import Any, Callable, Iterable, List, Mapping, Set, Union

import pystow
import requests
from cachier import cachier
from tqdm.auto import tqdm

from ..struct import Term

ICD_CLIENT_ID = pystow.get_config("pyobo", "icd_client_id")
ICD_CLIENT_SECRET = pystow.get_config("pyobo", "icd_client_secret")

TOKEN_URL = "https://icdaccessmanagement.who.int/connect/token"  # noqa:S105

ICD_BASE_URL = "https://id.who.int/icd"

ICD11_TOP_LEVEL_URL = f"{ICD_BASE_URL}/entity"
ICD10_TOP_LEVEL_URL = f"{ICD_BASE_URL}/release/10/2016"


def get_icd(url: str) -> requests.Response:
    """Get an ICD API endpoint."""
    return requests.get(url, headers=get_icd_api_headers())


def _get_entity(endpoint: str, identifier: str):
    url = f"{endpoint}/{identifier}"
    # tqdm.write(f'query {identifier} at {url}')
    res = get_icd(url)
    return res.json()


def get_child_identifiers(endpoint: str, res_json: Mapping[str, Any]) -> List[str]:
    """Ge the child identifiers."""
    return [url[len(endpoint) :].lstrip("/") for url in res_json.get("child", [])]


@cachier(stale_after=datetime.timedelta(minutes=45))
def get_icd_api_headers() -> Mapping[str, str]:
    """Get the headers, and refresh every hour."""
    grant_type = "client_credentials"
    body_params = {"grant_type": grant_type}
    tqdm.write("getting ICD API token")
    res = requests.post(TOKEN_URL, data=body_params, auth=(ICD_CLIENT_ID, ICD_CLIENT_SECRET))
    res_json = res.json()
    access_type = res_json["token_type"]
    access_token = res_json["access_token"]
    return {
        "API-Version": "v2",
        "Accept-Language": "en",
        "Authorization": f"{access_type} {access_token}",
    }


def visiter(
    identifier: str,
    visited_identifiers: Set[str],
    directory: Union[str, Path],
    *,
    endpoint: str,
    converter: Callable[[Mapping[str, Any]], Term],
) -> Iterable[Term]:
    """Iterate over all terms from the ICD endpoint."""
    path = Path(directory).joinpath(identifier).with_suffix(".json")
    if identifier in visited_identifiers:
        return
    visited_identifiers.add(identifier)

    if os.path.exists(path):
        with open(path) as file:
            res_json = json.load(file)
    else:
        res_json = _get_entity(endpoint, identifier)
        with open(path, "w") as file:
            json.dump(res_json, file, indent=2)

    yield converter(res_json)
    for identifier in get_child_identifiers(endpoint, res_json):
        yield from visiter(
            identifier,
            visited_identifiers,
            directory,
            converter=converter,
            endpoint=endpoint,
        )



# -*- coding: utf-8 -*-

"""Import ComPath mappings between pathways."""

from typing import Iterable

import pandas as pd
from pystow.utils import get_commit

from pyobo.constants import (
    PROVENANCE,
    SOURCE_ID,
    SOURCE_PREFIX,
    TARGET_ID,
    TARGET_PREFIX,
    XREF_COLUMNS,
)

__all__ = [
    "iter_compath_dfs",
]


def _get_df(name: str, *, sha: str, sep: str = ",") -> pd.DataFrame:
    url = f"https://raw.githubusercontent.com/ComPath/compath-resources/{sha}/mappings/{name}"
    df = pd.read_csv(
        url,
        sep=sep,
        usecols=["Source Resource", "Source ID", "Mapping Type", "Target Resource", "Target ID"],
    )
    df.rename(
        columns={
            "Source Resource": SOURCE_PREFIX,
            "Source ID": SOURCE_ID,
            "Target Resource": TARGET_PREFIX,
            "Target ID": TARGET_ID,
        },
        inplace=True,
    )
    df = df[df["Mapping Type"] == "equivalentTo"]
    del df["Mapping Type"]
    df[PROVENANCE] = url
    df = df[XREF_COLUMNS]

    df[SOURCE_PREFIX] = df[SOURCE_PREFIX].map(_fix_kegg_prefix)
    df[TARGET_PREFIX] = df[TARGET_PREFIX].map(_fix_kegg_prefix)
    df[SOURCE_ID] = [
        _fix_kegg_identifier(prefix, identifier)
        for prefix, identifier in df[[SOURCE_PREFIX, SOURCE_ID]].values
    ]
    df[TARGET_ID] = [
        _fix_kegg_identifier(prefix, identifier)
        for prefix, identifier in df[[TARGET_PREFIX, TARGET_ID]].values
    ]

    return df


def _fix_kegg_identifier(prefix, identifier) -> str:
    if prefix == "kegg.pathway":
        return identifier[len("path:") :]
    return identifier


def _fix_kegg_prefix(s):
    return s if s != "kegg" else "kegg.pathway"


def iter_compath_dfs() -> Iterable[pd.DataFrame]:
    """Iterate over all ComPath mappings."""
    sha = get_commit("ComPath", "compath-resources")

    yield _get_df("kegg_reactome.csv", sha=sha)
    yield _get_df("kegg_wikipathways.csv", sha=sha)
    yield _get_df("pathbank_kegg.csv", sha=sha)
    yield _get_df("pathbank_reactome.csv", sha=sha)
    yield _get_df("pathbank_wikipathways.csv", sha=sha)
    yield _get_df("special_mappings.csv", sha=sha)
    yield _get_df("wikipathways_reactome.csv", sha=sha)


def get_compath_xrefs_df() -> pd.DataFrame:
    """Iterate over all ComPath mappings."""
    return pd.concat(iter_compath_dfs())



# -*- coding: utf-8 -*-

"""Loading of the NCBI Taxonomy names."""

import csv
import gzip
from functools import lru_cache
from pathlib import Path
from typing import Mapping, Optional, Union

import requests

__all__ = [
    "load_ncbitaxon",
    "get_ncbitaxon_id",
    "get_ncbitaxon_name",
]

HERE = Path(__file__).parent.resolve()
PATH = HERE.joinpath("ncbitaxon.tsv.gz")
URL = "http://purl.obolibrary.org/obo/ncbitaxon.json"
URI_PREFIX = "http://purl.obolibrary.org/obo/NCBITaxon_"


@lru_cache(maxsize=1)
def load_ncbitaxon() -> Mapping[str, str]:
    """Load the NCBI Taxonomy identifier to name map."""
    return ensure(url=URL, path=PATH, uri_prefix=URI_PREFIX)


@lru_cache(maxsize=1)
def load_ncbitaxon_reverse() -> Mapping[str, str]:
    """Load the NCBI Taxonomy name to identifier map."""
    return {name: identifier for identifier, name in load_ncbitaxon().items()}


def get_ncbitaxon_name(ncbitaxon_id: str) -> Optional[str]:
    """Get the name from the identifier."""
    return load_ncbitaxon().get(ncbitaxon_id)


def get_ncbitaxon_id(name: str) -> Optional[str]:
    """Get the identifier from the name."""
    return load_ncbitaxon_reverse().get(name)


def ensure(url: str, path: Union[str, Path], uri_prefix: str) -> Mapping[str, str]:
    """Download the latest version of the resource."""
    path = Path(path)
    if path.is_file():
        with gzip.open(path, mode="rt") as file:
            return dict(csv.reader(file, delimiter="\t"))

    res_json = requests.get(url).json()

    rows = []
    for node in res_json["graphs"][0]["nodes"]:
        identifier = node["id"]
        if not identifier.startswith(uri_prefix):
            continue
        identifier = identifier[len(uri_prefix) :]
        name = node.get("lbl")
        if name:
            rows.append((identifier, name))

    with gzip.open(path, "wt") as file:
        writer = csv.writer(file, delimiter="\t")
        writer.writerows(sorted(rows))

    return dict(rows)


if __name__ == "__main__":
    load_ncbitaxon()



# -*- coding: utf-8 -*-

"""Extract registry information."""

from .metaregistry import (  # noqa: F401
    curie_has_blacklisted_prefix,
    curie_has_blacklisted_suffix,
    curie_is_blacklisted,
    get_remappings_full,
    get_remappings_prefix,
    get_xrefs_blacklist,
    get_xrefs_prefix_blacklist,
    get_xrefs_suffix_blacklist,
    has_no_download,
    iter_cached_obo,
    remap_full,
    remap_prefix,
)



# -*- coding: utf-8 -*-

"""Interface for caching data on AWS S3."""

import logging
import os
from typing import Optional, Set

import boto3
import humanize
from tabulate import tabulate

from pyobo import (
    get_id_name_mapping,
    get_id_synonyms_mapping,
    get_id_to_alts,
    get_properties_df,
    get_relations_df,
    get_xrefs_df,
)
from pyobo.api.utils import get_version
from pyobo.constants import RAW_DIRECTORY
from pyobo.registries import iter_cached_obo
from pyobo.utils.path import prefix_cache_join

__all__ = [
    "download_artifacts",
    "upload_artifacts",
    "upload_artifacts_for_prefix",
    "list_artifacts",
]

logger = logging.getLogger(__name__)


def download_artifacts(bucket: str, suffix: Optional[str] = None) -> None:
    """Download compiled parts from AWS.

    :param bucket: The name of the S3 bucket to download
    :param suffix: If specified, only download files with this suffix. Might
     be useful to specify ``suffix='names.tsv`` if you just want to run the
     name resolution service.
    """
    s3_client = boto3.client("s3")
    all_objects = s3_client.list_objects(Bucket=bucket)
    for entry in all_objects["Contents"]:
        key = entry["Key"]
        if suffix and not key.endswith(suffix):
            pass
        path = os.path.join(RAW_DIRECTORY, key)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        if os.path.exists(path):
            continue  # no need to download again
        logging.warning("downloading %s to %s", key, path)
        s3_client.download_file(bucket, key, path)


def upload_artifacts(
    bucket: str,
    whitelist: Optional[Set[str]] = None,
    blacklist: Optional[Set[str]] = None,
    s3_client=None,
) -> None:
    """Upload all artifacts to AWS."""
    if s3_client is None:
        s3_client = boto3.client("s3")
    all_objects = s3_client.list_objects(Bucket=bucket)
    uploaded_prefixes = {entry["Key"].split("/")[0] for entry in all_objects["Contents"]}

    for prefix, _ in sorted(iter_cached_obo()):
        if prefix in uploaded_prefixes:
            continue
        if whitelist and prefix not in whitelist:
            continue
        if blacklist and prefix in blacklist:
            continue
        upload_artifacts_for_prefix(prefix=prefix, bucket=bucket, s3_client=s3_client)


def upload_artifacts_for_prefix(*, prefix: str, bucket: str, s3_client=None):
    """Upload compiled parts for the given prefix to AWS."""
    if s3_client is None:
        s3_client = boto3.client("s3")

    logger.info("[%s] getting id->name mapping", prefix)
    get_id_name_mapping(prefix)
    id_name_path = prefix_cache_join(prefix, name="names.tsv", version=get_version(prefix))
    if not id_name_path.exists():
        raise FileNotFoundError
    id_name_key = os.path.join(prefix, "cache", "names.tsv")
    logger.info("[%s] uploading id->name mapping", prefix)
    upload_file(path=id_name_path, bucket=bucket, key=id_name_key, s3_client=s3_client)

    logger.info("[%s] getting id->synonyms mapping", prefix)
    get_id_synonyms_mapping(prefix)
    id_synonyms_path = prefix_cache_join(prefix, name="synonyms.tsv", version=get_version(prefix))
    if not id_synonyms_path.exists():
        raise FileNotFoundError
    id_synonyms_key = os.path.join(prefix, "cache", "synonyms.tsv")
    logger.info("[%s] uploading id->synonyms mapping", prefix)
    upload_file(path=id_synonyms_path, bucket=bucket, key=id_synonyms_key, s3_client=s3_client)

    logger.info("[%s] getting xrefs", prefix)
    get_xrefs_df(prefix)
    xrefs_path = prefix_cache_join(prefix, name="xrefs.tsv", version=get_version(prefix))
    if not xrefs_path.exists():
        raise FileNotFoundError
    xrefs_key = os.path.join(prefix, "cache", "xrefs.tsv")
    logger.info("[%s] uploading xrefs", prefix)
    upload_file(path=xrefs_path, bucket=bucket, key=xrefs_key, s3_client=s3_client)

    logger.info("[%s] getting relations", prefix)
    get_relations_df(prefix)
    relations_path = prefix_cache_join(prefix, name="relations.tsv", version=get_version(prefix))
    if not relations_path.exists():
        raise FileNotFoundError
    relations_key = os.path.join(prefix, "cache", "relations.tsv")
    logger.info("[%s] uploading relations", prefix)
    upload_file(path=relations_path, bucket=bucket, key=relations_key, s3_client=s3_client)

    logger.info("[%s] getting properties", prefix)
    get_properties_df(prefix)
    properties_path = prefix_cache_join(prefix, name="properties.tsv", version=get_version(prefix))
    if not properties_path.exists():
        raise FileNotFoundError
    properties_key = os.path.join(prefix, "cache", "properties.tsv")
    logger.info("[%s] uploading properties", prefix)
    upload_file(path=properties_path, bucket=bucket, key=properties_key, s3_client=s3_client)

    logger.info("[%s] getting alternative identifiers", prefix)
    get_id_to_alts(prefix)
    alts_path = prefix_cache_join(prefix, name="alt_ids.tsv", version=get_version(prefix))
    if not alts_path.exists():
        raise FileNotFoundError
    alts_key = os.path.join(prefix, "cache", "alt_ids.tsv")
    logger.info("[%s] uploading alternative identifiers", prefix)
    upload_file(path=alts_path, bucket=bucket, key=alts_key)


def upload_file(*, path, bucket, key, s3_client=None):
    """Upload a file to an S3 bucket.

    :param path: The local file path
    :param bucket: The name of the S3 bucket
    :param key: The relative file path to put on the S3 bucket
    """
    if s3_client is None:
        s3_client = boto3.client("s3")
    s3_client.upload_file(path, bucket, key)


def list_artifacts(bucket: str) -> str:
    """List the files in a given bucket."""
    s3_client = boto3.client("s3")
    all_objects = s3_client.list_objects(Bucket=bucket)
    rows = [
        (entry["Key"], humanize.naturalsize(entry["Size"])) for entry in all_objects["Contents"]
    ]
    return tabulate(rows, headers=["File", "Size"])



# -*- coding: utf-8 -*-

"""Get Wikidata xrefs.

Run with ``python -m pyobo.xrefdb.sources.wikidata``.
"""

import json
import logging
from typing import Iterable, Tuple

import bioregistry
import click
import pandas as pd
import requests
from more_click import verbose_option
from tqdm.auto import tqdm

from ...constants import RAW_MODULE, XREF_COLUMNS
from ...version import get_version

logger = logging.getLogger(__name__)

#: WikiData SPARQL endpoint. See https://www.wikidata.org/wiki/Wikidata:SPARQL_query_service#Interfacing
URL = "https://query.wikidata.org/bigdata/namespace/wdq/sparql"

WIKIDATA_MAPPING_DIRECTORY = RAW_MODULE.submodule("wikidata", "mappings")


def get_wikidata_xrefs_df(*, use_tqdm: bool = True) -> pd.DataFrame:
    """Get all Wikidata xrefs."""
    return pd.concat(iterate_wikidata_dfs(use_tqdm=use_tqdm))


def iterate_wikidata_dfs(*, use_tqdm: bool = True) -> Iterable[pd.DataFrame]:
    """Iterate over WikiData xref dataframes."""
    wikidata_properties = {
        prefix: entry.wikidata["prefix"]
        for prefix, entry in bioregistry.read_registry().items()
        if entry.wikidata and "prefix" in entry.wikidata
    }

    it = tqdm(sorted(wikidata_properties.items()), disable=not use_tqdm, desc="Wikidata properties")
    for prefix, wikidata_property in it:
        if prefix in {"pubmed", "pmc", "orcid", "inchi", "smiles"}:
            continue  # too many
        it.set_postfix({"prefix": prefix})
        try:
            yield get_wikidata_df(prefix, wikidata_property)
        except json.decoder.JSONDecodeError as e:
            logger.warning(
                "[%s] Problem decoding results from %s: %s", prefix, wikidata_property, e
            )


def get_wikidata_df(prefix: str, wikidata_property: str) -> pd.DataFrame:
    """Get Wikidata xrefs."""
    df = pd.DataFrame(
        [
            ("wikidata", wikidata_id, prefix, external_id, "wikidata")
            for wikidata_id, external_id in iter_wikidata_mappings(wikidata_property)
        ],
        columns=XREF_COLUMNS,
    )
    logger.debug("got wikidata (%s; %s): %d rows", prefix, wikidata_property, len(df.index))
    return df


def iter_wikidata_mappings(wikidata_property: str) -> Iterable[Tuple[str, str]]:
    """Iterate over Wikidata xrefs."""
    path = WIKIDATA_MAPPING_DIRECTORY.join(name=f"{wikidata_property}.json")
    if path.exists():
        with path.open() as file:
            rows = json.load(file)
    else:
        query = f"SELECT ?wikidata_id ?id WHERE {{?wikidata_id wdt:{wikidata_property} ?id}}"
        rows = _run_query(query)
        with path.open("w") as file:
            json.dump(rows, file, indent=2)

    for row in rows:
        wikidata_id = row["wikidata_id"]["value"][len("http://wikidata.org/entity/") :]
        entity_id = row["id"]["value"]
        yield wikidata_id, entity_id


HEADERS = {
    "User-Agent": f"pyobo/{get_version()}",
}


def _run_query(query, base: str = URL):
    logger.debug("running query: %s", query)
    res = requests.get(base, params={"query": query, "format": "json"}, headers=HEADERS)
    res.raise_for_status()
    res_json = res.json()
    return res_json["results"]["bindings"]


@click.command()
@verbose_option
def _main():
    """Summarize xrefs."""
    for _ in iterate_wikidata_dfs():
        pass


if __name__ == "__main__":
    _main()



# -*- coding: utf-8 -*-

"""Converter for ChEBI."""

from typing import Mapping, Set, Tuple

from ..api import get_filtered_properties_mapping, get_filtered_relations_df
from ..struct import Reference, TypeDef
from ..utils.io import multisetdict

__all__ = [
    "get_chebi_smiles_id_mapping",
    "get_chebi_id_smiles_mapping",
    "get_chebi_role_to_children",
]


def get_chebi_id_smiles_mapping() -> Mapping[str, str]:
    """Get a mapping from ChEBI identifiers to SMILES.

    This is common enough that it gets its own function :)
    """
    return get_filtered_properties_mapping("chebi", "http://purl.obolibrary.org/obo/chebi/smiles")


def get_chebi_smiles_id_mapping() -> Mapping[str, str]:
    """Get a mapping from sSMILES to ChEBI identifiers."""
    return {v: k for k, v in get_chebi_id_smiles_mapping().items()}


has_role = TypeDef(reference=Reference(prefix="chebi", identifier="has_role"))


def get_chebi_role_to_children() -> Mapping[str, Set[Tuple[str, str]]]:
    """Get the ChEBI role to children mapping."""
    df = get_filtered_relations_df("chebi", relation=has_role)
    return multisetdict((role_id, ("chebi", chemical_id)) for chemical_id, _, role_id in df.values)



"""Convert PyOBO into OBO Graph."""

from typing import Iterable, List

import bioregistry
import curies
from bioontologies.obograph import (
    OBO_SYNONYM_TO_OIO,
    OIO_TO_REFERENCE,
    Definition,
    Edge,
    Graph,
    GraphDocument,
    Meta,
    Node,
    Synonym,
    Xref,
)
from bioontologies.robot import ParseResults

from pyobo.struct import Obo, Reference, Term
from pyobo.struct.typedef import is_a

__all__ = [
    "graph_from_obo",
    "parse_results_from_obo",
]


def parse_results_from_obo(obo: Obo) -> ParseResults:
    """Get parse results from an OBO graph."""
    graph = graph_from_obo(obo)
    return ParseResults(graph_document=GraphDocument(graphs=[graph]))


def graph_from_obo(obo: Obo) -> Graph:
    """Get an OBO Graph object from a PyOBO object."""
    nodes: List[Node] = []
    edges: List[Edge] = []
    for term in obo:
        nodes.append(_get_class_node(term))
        edges.extend(_iter_edges(term))
    return Graph(
        id=bioregistry.get_bioregistry_iri("bioregistry", obo.ontology),
        prefix=obo.ontology,
        meta=_get_meta(obo),
        nodes=nodes,
        edges=edges,
        standardized=True,  # from construction :)
    )


def _get_meta(obo: Obo) -> Meta:
    return Meta(
        version=obo.data_version,
    )


def _rewire(r: Reference) -> curies.Reference:
    return curies.Reference(prefix=r.prefix, identifier=r.identifier)


def _get_class_node(term: Term) -> Node:
    if not term.definition:
        definition = None
    else:
        definition = Definition.from_parsed(
            value=term.definition, references=[_rewire(p) for p in term.provenance]
        )

    if term.xrefs:
        if not term.xref_types:
            term.xref_types = [
                Reference(prefix="oboInOwl", identifier="hasDbXref") for _ in term.xrefs
            ]
        elif len(term.xrefs) != len(term.xref_types):
            raise ValueError

    xrefs = [
        Xref.from_parsed(
            predicate=_rewire(xref_type),
            value=_rewire(xref),
        )
        for xref, xref_type in zip(term.xrefs, term.xref_types)
    ]
    default_st = Reference(prefix="oboInOwl", identifier="SynonymType")
    synonyms = [
        Synonym.from_parsed(
            name=synonym.name,
            predicate=OIO_TO_REFERENCE[OBO_SYNONYM_TO_OIO[synonym.specificity]],
            synonym_type=_rewire(synonym.type.reference) if synonym.type else default_st,
            references=[_rewire(x) for x in synonym.provenance],
        )
        for synonym in term.synonyms
    ]

    meta = Meta(
        definition=definition,
        xrefs=xrefs,
        synonyms=synonyms,
        basicPropertyValues=None,  # TODO properties
        deprecated=term.is_obsolete or False,
    )
    return Node(
        id=term.bioregistry_link,
        lbl=term.name,
        meta=meta,
        type="CLASS",
        reference=_rewire(term.reference),
        standardized=True,
    )


def _iter_edges(term: Term) -> Iterable[Edge]:
    for parent in term.parents:
        yield Edge.from_parsed(
            _rewire(term.reference),
            _rewire(is_a.reference),
            _rewire(parent),
        )

    for typedef, targets in term.relationships.items():
        for target in targets:
            yield Edge.from_parsed(
                _rewire(term.reference),
                _rewire(typedef.reference),
                _rewire(target),
            )



# -*- coding: utf-8 -*-

"""Converter for dictyBase gene.

Note that normal dictybase idenififers are for sequences
"""

import logging
from typing import Iterable

import pandas as pd
from tqdm.auto import tqdm

from pyobo.struct import Obo, Reference, Synonym, Term, from_species, has_gene_product
from pyobo.utils.io import multisetdict
from pyobo.utils.path import ensure_df

__all__ = [
    "DictybaseGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "dictybase.gene"
URL = (
    "http://dictybase.org/db/cgi-bin/dictyBase/download/"
    "download.pl?area=general&ID=gene_information.txt"
)
UNIPROT_MAPPING = (
    "http://dictybase.org/db/cgi-bin/dictyBase/download/"
    "download.pl?area=general&ID=DDB-GeneID-UniProt.txt"
)


class DictybaseGetter(Obo):
    """An ontology representation of Dictybase's gene nomenclature."""

    ontology = PREFIX
    dynamic_version = True
    typedefs = [from_species, has_gene_product]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force)


def get_obo(force: bool = False) -> Obo:
    """Get dictyBase Gene as OBO."""
    return DictybaseGetter(force=force)


def get_terms(force: bool = False) -> Iterable[Term]:
    """Get terms."""
    # DDB ID	DDB_G ID	Name	UniProt ID
    uniprot_mappings = multisetdict(
        ensure_df(PREFIX, url=URL, force=force, name="uniprot_mappings.tsv", usecols=[1, 3]).values
    )

    terms = ensure_df(PREFIX, url=URL, force=force, name="gene_info.tsv")
    # GENE ID (DDB_G ID)	Gene Name	Synonyms	Gene products
    for identifier, name, synonyms, products in tqdm(terms.values):
        term = Term.from_triple(
            prefix=PREFIX,
            identifier=identifier,
            name=name,
        )
        if products and pd.notna(products) and products != "unknown":
            for synonym in products.split(","):
                term.append_synonym(synonym.strip())
        if synonyms and pd.notna(synonyms):
            for synonym in synonyms.split(","):
                term.append_synonym(Synonym(synonym.strip()))
        for uniprot_id in uniprot_mappings.get(identifier, []):
            if not uniprot_id or pd.isna(uniprot_id) or uniprot_id not in {"unknown", "pseudogene"}:
                continue
            term.append_relationship(has_gene_product, Reference.auto("uniprot", uniprot_id))

        term.set_species(identifier="44689", name="Dictyostelium discoideum")
        yield term


if __name__ == "__main__":
    DictybaseGetter.cli()



# -*- coding: utf-8 -*-

"""Tests for alternative identifiers."""

import unittest

from pyobo import get_name, get_name_by_curie, get_primary_curie, get_primary_identifier
from pyobo.mocks import get_mock_id_alts_mapping, get_mock_id_name_mapping

mock_id_alts_mapping = get_mock_id_alts_mapping(
    {
        "go": {
            "0003700": ["0001071"],
        },
    }
)

mock_id_names_mapping = get_mock_id_name_mapping(
    {
        "go": {
            "0003700": "DNA-binding transcription factor activity",
        },
        "ncbitaxon": {
            "52818": "Allamanda cathartica",
        },
    }
)


class TestAltIds(unittest.TestCase):
    """Tests for alternative identifiers."""

    @mock_id_alts_mapping
    @mock_id_names_mapping
    def test_get_primary(self, _, __):
        """Test upgrading an obsolete identifier."""
        primary_id = get_primary_identifier("go", "0001071")
        self.assertIsNotNone(primary_id)
        self.assertEqual("0003700", primary_id)
        name = get_name("go", "0001071")
        self.assertIsNotNone(name)
        self.assertEqual("DNA-binding transcription factor activity", name)

    @mock_id_alts_mapping
    @mock_id_names_mapping
    def test_get_primary_by_curie(self, _, __):
        """Test upgrading an obsolete CURIE."""
        primary_curie = get_primary_curie("go:0001071")
        self.assertIsNotNone(primary_curie)
        self.assertEqual("go:0003700", primary_curie)
        name = get_name_by_curie("go:0001071")
        self.assertIsNotNone(name)
        self.assertEqual("DNA-binding transcription factor activity", name)

    @mock_id_alts_mapping
    @mock_id_names_mapping
    def test_already_primary(self, _, __):
        """Test when you give a primary id."""
        primary_id = get_primary_identifier("go", "0003700")
        self.assertIsNotNone(primary_id)
        self.assertEqual("0003700", primary_id)
        name = get_name("go", "0003700")
        self.assertIsNotNone(name)
        self.assertEqual("DNA-binding transcription factor activity", name)

    @mock_id_alts_mapping
    @mock_id_names_mapping
    def test_already_primary_by_curie(self, _, __):
        """Test when you give a primary CURIE."""
        primary_curie = get_primary_curie("go:0003700")
        self.assertIsNotNone(primary_curie)
        self.assertEqual("go:0003700", primary_curie)
        name = get_name_by_curie("go:0003700")
        self.assertIsNotNone(name)
        self.assertEqual("DNA-binding transcription factor activity", name)

    @mock_id_alts_mapping
    @mock_id_names_mapping
    def test_no_alts(self, _, __):
        """Test alternate behavior for nomenclature source with no alts."""
        primary_id = get_primary_identifier("ncbitaxon", "52818")
        self.assertEqual("52818", primary_id)
        self.assertEqual("Allamanda cathartica", get_name("ncbitaxon", "52818"))



# -*- coding: utf-8 -*-

"""Get ChEMBL xrefs."""

from typing import Optional

import bioversions
import pandas as pd

from pyobo.constants import (
    PROVENANCE,
    SOURCE_ID,
    SOURCE_PREFIX,
    TARGET_ID,
    TARGET_PREFIX,
    XREF_COLUMNS,
)
from pyobo.utils.path import ensure_df

CHEMBL_COMPOUND_PREFIX = "chembl.compound"
CHEMBL_TARGET_PREFIX = "chembl.target"


def get_chembl_compound_equivalences_raw(
    usecols=None, version: Optional[str] = None
) -> pd.DataFrame:
    """Get the chemical representations raw dataframe."""
    if version is None:
        version = bioversions.get_version("chembl")

    base_url = f"ftp://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_{version}"
    url = f"{base_url}/chembl_{version}_chemreps.txt.gz"
    return ensure_df(CHEMBL_COMPOUND_PREFIX, url=url, sep="\t", usecols=usecols)


def get_chembl_compound_equivalences(version: Optional[str] = None) -> pd.DataFrame:
    """Get ChEMBL chemical equivalences."""
    if version is None:
        version = bioversions.get_version("chembl")

    df = get_chembl_compound_equivalences_raw(version=version)
    rows = []
    for chembl, _smiles, _inchi, inchi_key in df.values:
        rows.extend(
            [
                # No smiles/inchi since they can have variable length
                # ("chembl.compound", chembl, "smiles", smiles, f"chembl{version}"),
                # ("chembl.compound", chembl, "inchi", inchi, f"chembl{version}"),
                ("chembl.compound", chembl, "inchikey", inchi_key, f"chembl{version}"),
            ]
        )
    return pd.DataFrame(rows, columns=XREF_COLUMNS)


def get_chembl_protein_equivalences(version: Optional[str] = None) -> pd.DataFrame:
    """Get ChEMBL protein equivalences."""
    if version is None:
        version = bioversions.get_version("chembl")

    url = f"ftp://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_{version}/chembl_uniprot_mapping.txt"
    df = ensure_df(
        CHEMBL_TARGET_PREFIX,
        url=url,
        sep="\t",
        usecols=[0, 1],
        names=[TARGET_ID, SOURCE_ID],  # switch around
    )
    df.loc[:, SOURCE_PREFIX] = "chembl.target"
    df.loc[:, TARGET_PREFIX] = "uniprot"
    df.loc[:, PROVENANCE] = f"chembl{version}"
    df = df[XREF_COLUMNS]
    return df


def get_chembl_xrefs_df(version: Optional[str] = None) -> pd.DataFrame:
    """Get all ChEBML equivalences."""
    if version is None:
        version = bioversions.get_version("chembl")

    return pd.concat(
        [
            get_chembl_compound_equivalences(version=version),
            get_chembl_protein_equivalences(version=version),
        ]
    )



# -*- coding: utf-8 -*-

"""Converter for NPASS."""

import logging
from typing import Iterable

import pandas as pd
from tqdm.auto import tqdm

from ..struct import Obo, Reference, Synonym, Term
from ..utils.path import ensure_df

__all__ = [
    "NPASSGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "npass"


# TODO add InChI, InChI-key, and SMILES information from NPASS, if desired
# METADATA_URL = f'{BASE_URL}_naturalProducts_properties.txt'


class NPASSGetter(Obo):
    """An ontology representation of NPASS's chemical nomenclature."""

    ontology = bioversions_key = PREFIX

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get NPASS as OBO."""
    return NPASSGetter()


def get_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get the NPASS chemical nomenclature."""
    base_url = f"http://bidd.group/NPASS/downloadFiles/NPASSv{version}_download"
    url = f"{base_url}_naturalProducts_generalInfo.txt"
    return ensure_df(
        PREFIX,
        url=url,
        version=version,
        dtype=str,
        encoding="ISO-8859-1",
        na_values={"NA", "n.a.", "nan"},
        force=force,
    )


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate NPASS terms."""
    df = get_df(version=version, force=force)
    it = tqdm(df.values, total=len(df.index), desc=f"mapping {PREFIX}")
    for identifier, name, iupac, chembl_id, pubchem_compound_ids, _, _, _, _ in it:
        term = Term.from_triple(
            PREFIX, identifier=identifier, name=name if pd.notna(name) else identifier
        )

        for xref_prefix, xref_id in [
            ("chembl", chembl_id),
            # ("zinc", zinc_id),
        ]:
            if pd.notna(xref_id):
                term.append_xref(Reference(prefix=xref_prefix, identifier=xref_id))

        # TODO check that the first is always the parent compound?
        if pd.notna(pubchem_compound_ids):
            pubchem_compound_ids = pubchem_compound_ids.split(";")
            if len(pubchem_compound_ids) > 1:
                logger.debug("multiple cids for %s: %s", identifier, pubchem_compound_ids)
            for pubchem_compound_id in pubchem_compound_ids:
                term.append_xref(
                    Reference(prefix="pubchem.compound", identifier=pubchem_compound_id)
                )

        for synonym in [iupac]:
            if pd.notna(synonym):
                term.append_synonym(Synonym(name=synonym))

        yield term


if __name__ == "__main__":
    NPASSGetter.cli()



# -*- coding: utf-8 -*-

"""Converter for CONSO."""

from typing import Iterable, List

import pandas as pd

from ..struct import Obo, Reference, Synonym, Term
from ..utils.io import multidict
from ..utils.path import ensure_df

__all__ = [
    "CONSOGetter",
]

PREFIX = "conso"
BASE_URL = "https://raw.githubusercontent.com/pharmacome/conso/master/src/conso/resources"
TERMS_URL = f"{BASE_URL}/terms.tsv"
RELATIONS_URL = f"{BASE_URL}/relations.tsv"
TYPEDEFS_URL = f"{BASE_URL}/typedefs.tsv"
SYNONYMS_URL = f"{BASE_URL}/synonyms.tsv"


class CONSOGetter(Obo):
    """An ontology representation of CONSO vocabulary."""

    ontology = PREFIX
    dynamic_version = True

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms()


def get_obo() -> Obo:
    """Get CONSO as OBO."""
    return CONSOGetter()


def iter_terms() -> Iterable[Term]:
    """Get CONSO terms."""
    terms_df = ensure_df(PREFIX, url=TERMS_URL)

    synonyms_df = ensure_df(PREFIX, url=SYNONYMS_URL)
    synonyms_df["reference"] = synonyms_df["reference"].map(
        lambda s: [Reference.from_curie(s)] if pd.notna(s) and s != "?" else [],
    )
    synonyms_df["specificity"] = synonyms_df["specificity"].map(
        lambda s: "EXACT" if pd.isna(s) or s == "?" else s
    )

    synonyms = multidict(
        (
            identifier,
            Synonym(
                name=synonym,
                provenance=provenance,
                specificity=specificity,
            ),
        )
        for identifier, synonym, provenance, specificity in synonyms_df.values
    )

    # TODO later
    # relations_df = ensure_df(PREFIX, RELATIONS_URL)

    for _, row in terms_df.iterrows():
        if row["Name"] == "WITHDRAWN":
            continue
        provenance: List[Reference] = []
        for curie in row["References"].split(","):
            curie = curie.strip()
            if not curie:
                continue
            reference = Reference.from_curie(curie)
            if reference is not None:
                provenance.append(reference)
        identifier = row["Identifier"]
        yield Term(
            reference=Reference(prefix=PREFIX, identifier=identifier, name=row["Name"]),
            definition=row["Description"],
            provenance=provenance,
            synonyms=synonyms.get(identifier, []),
        )


if __name__ == "__main__":
    CONSOGetter.cli()



# -*- coding: utf-8 -*-

"""Mocks for PyOBO."""

from typing import List, Mapping, Optional, Tuple, TypeVar, Union
from unittest import mock

import pandas as pd

from pyobo.constants import XREF_COLUMNS

__all__ = [
    "get_mock_id_name_mapping",
    "get_mock_id_synonyms_mapping",
    "get_mock_get_xrefs_df",
    "get_mock_id_alts_mapping",
]


def get_mock_id_name_mapping(data: Mapping[str, Mapping[str, str]]) -> mock._patch:
    """Mock the :func:`pyobo.extract.get_id_name_mapping` function.

    :param data: A mapping from prefix to mappings of identifier to names.
    """
    return _replace_mapping_getter("pyobo.api.names.get_id_name_mapping", data)


def get_mock_id_synonyms_mapping(data: Mapping[str, Mapping[str, List[str]]]) -> mock._patch:
    """Mock the :func:`pyobo.extract.get_id_synonyms_mapping` function.

    :param data: A mapping from prefix to mappings of identifier to lists of synonyms.
    """
    return _replace_mapping_getter("pyobo.api.names.get_id_synonyms_mapping", data)


def get_mock_id_alts_mapping(data: Mapping[str, Mapping[str, List[str]]]) -> mock._patch:
    """Mock the :func:`pyobo.extract.get_id_to_alts` function.

    :param data: A mapping from prefix to mappings of identifier to lists of alternative identifiers.
    """
    return _replace_mapping_getter("pyobo.api.alts.get_id_to_alts", data)


X = TypeVar("X")


def _replace_mapping_getter(name: str, data: Mapping[str, Mapping[str, X]]) -> mock._patch:
    def _mock_get_data(prefix: str, **_kwargs) -> Mapping[str, X]:
        return data.get(prefix, {})

    return mock.patch(name, side_effect=_mock_get_data)


def get_mock_get_xrefs_df(
    df: Union[List[Tuple[str, str, str, str, str]], pd.DataFrame]
) -> mock._patch:
    """Mock the :func:`pyobo.xrefsdb.xrefs_pipeline.get_xref_df` function.

    :param df: The dataframe to return when the function is called
    """
    if isinstance(df, list):
        df = pd.DataFrame(df, columns=XREF_COLUMNS)

    def _mock_get_xrefs_df(*_args, **_kwargs) -> pd.DataFrame:
        return df

    return mock.patch(
        "pyobo.resource_utils.ensure_inspector_javert_df", side_effect=_mock_get_xrefs_df
    )


def _make_mock_get_name(name: str, data: Mapping[str, Mapping[str, X]]) -> mock._patch:
    def _get_name(prefix: str, identifier: str) -> Optional[X]:
        return data.get(prefix, {}).get(identifier)

    return mock.patch(name, side_effect=_get_name)



# -*- coding: utf-8 -*-

"""PyOBO's Gilda Service."""

from typing import Iterable, Union

import flask
from flask_bootstrap import Bootstrap
from flask_wtf import FlaskForm
from wtforms.fields import StringField, SubmitField
from wtforms.validators import DataRequired

from pyobo.gilda_utils import get_grounder


class Form(FlaskForm):
    """Form for submitting a query."""

    text = StringField("Text", validators=[DataRequired()])
    submit = SubmitField()

    def make_response(self):
        """Make a response with the text."""
        return flask.redirect(flask.url_for("ground", text=self.text.data))


def get_app(prefix: Union[str, Iterable[str]]):
    """Make an app for grounding the text."""
    grounder = get_grounder(prefix)

    app = flask.Flask(__name__)
    app.config["WTF_CSRF_ENABLED"] = False
    Bootstrap(app)

    @app.route("/", methods=["GET", "POST"])
    def home():
        """Ground the given text."""
        form = Form()
        if form.validate_on_submit():
            return form.make_response()
        return flask.render_template("home.html", form=form)

    @app.route("/ground/<text>")
    def ground(text: str):
        """Ground the given text."""
        return flask.jsonify([scored_match.to_json() for scored_match in grounder.ground(text)])

    return app



# -*- coding: utf-8 -*-

"""Data structures for OBO."""

from typing import Optional, Tuple

import bioregistry
import curies
from pydantic import Field, validator

from .utils import obo_escape
from ..identifier_utils import normalize_curie

__all__ = [
    "Reference",
    "Referenced",
]


class Reference(curies.Reference):
    """A namespace, identifier, and label."""

    name: Optional[str] = Field(description="the name of the reference")

    @validator("prefix")
    def validate_prefix(cls, v):  # noqa
        """Validate the prefix for this reference."""
        norm_prefix = bioregistry.normalize_prefix(v)
        if norm_prefix is None:
            raise ValueError(f"Unknown prefix: {v}")
        return norm_prefix

    @classmethod
    def auto(cls, prefix: str, identifier: str) -> "Reference":
        """Create a reference and auto-populate its name."""
        from ..api import get_name

        name = get_name(prefix, identifier)
        return cls(prefix=prefix, identifier=identifier, name=name)

    @property
    def bioregistry_link(self) -> str:
        """Get the bioregistry link."""
        return f"https://bioregistry.io/{self.curie}"

    @classmethod
    def from_curie(
        cls,
        curie: str,
        name: Optional[str] = None,
        *,
        strict: bool = True,
        auto: bool = False,
    ) -> Optional["Reference"]:
        """Get a reference from a CURIE.

        :param curie: The compact URI (CURIE) to parse in the form of `<prefix>:<identifier>`
        :param name: The name associated with the CURIE
        :param strict: If true, raises an error if the CURIE can not be parsed.
        :param auto: Automatically look up name
        """
        prefix, identifier = normalize_curie(curie, strict=strict)
        return cls._materialize(prefix=prefix, identifier=identifier, name=name, auto=auto)

    @classmethod
    def from_iri(
        cls,
        iri: str,
        name: Optional[str] = None,
        *,
        auto: bool = False,
    ) -> Optional["Reference"]:
        """Get a reference from an IRI using the Bioregistry.

        :param iri: The IRI to parse
        :param name: The name associated with the CURIE
        :param auto: Automatically look up name
        """
        prefix, identifier = bioregistry.parse_iri(iri)
        return cls._materialize(prefix=prefix, identifier=identifier, name=name, auto=auto)

    @classmethod
    def _materialize(
        cls,
        prefix: Optional[str],
        identifier: Optional[str],
        name: Optional[str] = None,
        *,
        auto: bool = False,
    ) -> Optional["Reference"]:
        if prefix is None or identifier is None:
            return None
        if name is None and auto:
            return cls.auto(prefix=prefix, identifier=identifier)
        return cls(prefix=prefix, identifier=identifier, name=name)

    @property
    def _escaped_identifier(self):
        return obo_escape(self.identifier)

    def __str__(self):  # noqa: D105
        identifier_lower = self.identifier.lower()
        if identifier_lower.startswith(f"{self.prefix.lower()}:"):
            rv = identifier_lower
        else:
            rv = f"{self.prefix}:{self._escaped_identifier}"
        if self.name:
            rv = f"{rv} ! {self.name}"
        return rv

    def __hash__(self):  # noqa: D105
        return hash((self.__class__, self.prefix, self.identifier))


class Referenced:
    """A class that contains a reference."""

    reference: Reference

    @property
    def prefix(self):
        """The prefix of the typedef."""  # noqa: D401
        return self.reference.prefix

    @property
    def name(self):
        """The name of the typedef."""  # noqa: D401
        return self.reference.name

    @property
    def identifier(self) -> str:
        """The local unique identifier for this typedef."""  # noqa: D401
        return self.reference.identifier

    @property
    def curie(self) -> str:
        """The CURIE for this typedef."""  # noqa: D401
        return self.reference.curie

    @property
    def pair(self) -> Tuple[str, str]:
        """The pair of namespace/identifier."""  # noqa: D401
        return self.reference.pair

    @property
    def bioregistry_link(self) -> str:
        """Get the bioregistry link."""
        return self.reference.bioregistry_link



# -*- coding: utf-8 -*-

"""High-level API for typedefs."""

import logging
from functools import lru_cache

import pandas as pd

from .utils import get_version
from ..getters import get_ontology
from ..identifier_utils import wrap_norm_prefix
from ..utils.cache import cached_df
from ..utils.path import prefix_cache_join

__all__ = [
    "get_typedef_df",
]

logger = logging.getLogger(__name__)


@lru_cache()
@wrap_norm_prefix
def get_typedef_df(prefix: str, force: bool = False) -> pd.DataFrame:
    """Get an identifier to name mapping for the typedefs in an OBO file."""
    version = get_version(prefix)
    path = prefix_cache_join(prefix, name="typedefs.tsv", version=version)

    @cached_df(path=path, dtype=str, force=force)
    def _df_getter() -> pd.DataFrame:
        logger.debug("[%s] no cached typedefs found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, version=version)
        logger.debug("[%s] loading typedef mappings", prefix)
        return ontology.get_typedef_df()

    return _df_getter()



# -*- coding: utf-8 -*-

"""Constants for tests for PyOBO."""

import pathlib
from unittest import mock

HERE = pathlib.Path(__file__).parent.resolve()
RESOURCES = HERE / "resources"

TEST_CHEBI_OBO_PATH = RESOURCES / "test_chebi.obo"
TEST_CHEBI_OBO_URL = pathlib.Path(TEST_CHEBI_OBO_PATH).as_uri()

TEST_GMT_PATH = RESOURCES / "test_msigdb.gmt"
TEST_WP_GMT_PATH = RESOURCES / "test_wikipathways.gmt"

chebi_patch = mock.patch(
    "pyobo.getters._ensure_ontology_path",
    side_effect=lambda *args, **kwargs: ("obo", TEST_CHEBI_OBO_PATH),
)



# -*- coding: utf-8 -*-

"""Converter for CVX."""

from collections import defaultdict
from typing import Iterable

import pandas as pd

from pyobo import Obo, Term

__all__ = [
    "CVXGetter",
]

cvx_url = "https://www2a.cdc.gov/vaccines/iis/iisstandards/downloads/cvx.txt"
PREFIX = "cvx"


class CVXGetter(Obo):
    """An ontology representation of CVX."""

    ontology = PREFIX
    dynamic_version = True

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms()


def iter_terms() -> Iterable[Term]:
    """Iterate over terms in CVX."""
    dd = defaultdict(set)
    hierarchy_df = pd.read_csv(
        "https://www2a.cdc.gov/vaccines/iis/iisstandards/downloads/VG.txt",
        sep="|",
        usecols=[1, 4],
        dtype=str,
    )
    for col in hierarchy_df.columns:
        hierarchy_df[col] = hierarchy_df[col].map(str.strip)
    for child, parent in hierarchy_df.values:
        dd[child].add(parent)

    cvx_df = pd.read_csv(
        cvx_url,
        sep="|",
        names=[
            "cvx",
            "short_name",
            "full_name",
            "notes",
            "status",
            "nonvaccine",
            "updated",
        ],
        dtype=str,
    )
    for col in cvx_df.columns:
        cvx_df[col] = cvx_df[col].map(lambda s: s.strip() if pd.notna(s) else s)
    terms = {}
    for cvx, short_name, full_name, notes, status, nonvaccine, _updated in cvx_df.values:
        term = Term.from_triple(PREFIX, cvx, full_name)
        if short_name != full_name:
            term.append_synonym(short_name)
        if pd.notna(notes):
            term.append_comment(notes)
        if pd.notna(status):
            term.append_property("status", status)
        if pd.notna(nonvaccine):
            term.append_property("nonvaccine", nonvaccine)
        terms[cvx] = term

    for child, parents in dd.items():
        for parent in sorted(parents):
            parent_term = terms[parent]
            terms[child].append_parent(parent_term)

    return iter(sorted(terms.values(), key=lambda term: int(term.identifier)))


if __name__ == "__main__":
    CVXGetter.cli()



# -*- coding: utf-8 -*-

"""Utilities for OBO files."""

import datetime
import gzip
import json
import logging
import pathlib
import subprocess
import typing
import urllib.error
from collections import Counter
from pathlib import Path
from typing import (
    Callable,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import bioregistry
from tqdm.auto import tqdm

from .constants import DATABASE_DIRECTORY
from .identifier_utils import MissingPrefix, wrap_norm_prefix
from .plugins import has_nomenclature_plugin, run_nomenclature_plugin
from .struct import Obo
from .utils.io import get_writer
from .utils.path import ensure_path, prefix_directory_join
from .version import get_git_hash, get_version

__all__ = [
    "get_ontology",
    "NoBuild",
]

logger = logging.getLogger(__name__)


class NoBuild(RuntimeError):
    """Base exception for being unable to build."""


class UnhandledFormat(NoBuild):
    """Only OWL is available."""


#: The following prefixes can not be loaded through ROBOT without
#: turning off integrity checks
REQUIRES_NO_ROBOT_CHECK = {"clo"}


@wrap_norm_prefix
def get_ontology(
    prefix: str,
    *,
    force: bool = False,
    rewrite: bool = False,
    strict: bool = True,
    version: Optional[str] = None,
    robot_check: bool = True,
) -> Obo:
    """Get the OBO for a given graph.

    :param prefix: The prefix of the ontology to look up
    :param version: The pre-looked-up version of the ontology
    :param force: Download the data again
    :param rewrite: Should the OBO cache be rewritten? Automatically set to true if ``force`` is true
    :param strict: Should CURIEs be treated strictly? If true, raises exceptions on invalid/malformed
    :param robot_check:
        If set to false, will send the ``--check=false`` command to ROBOT to disregard
        malformed ontology components. Necessary to load some ontologies like VO.
    :returns: An OBO object

    :raises OnlyOWLError: If the OBO foundry only has an OWL document for this resource.

    Alternate usage if you have a custom url::

    >>> from pystow.utils import download
    >>> from pyobo import Obo, from_obo_path
    >>> url = ...
    >>> obo_path = ...
    >>> download(url=url, path=path)
    >>> obo = from_obo_path(path)
    """
    if force:
        rewrite = True
    if prefix == "uberon":
        logger.info("UBERON has so much garbage in it that defaulting to non-strict parsing")
        strict = False

    obonet_json_gz_path = prefix_directory_join(
        prefix, name=f"{prefix}.obonet.json.gz", ensure_exists=False, version=version
    )
    if obonet_json_gz_path.exists() and not force:
        from .reader import from_obonet
        from .utils.cache import get_gzipped_graph

        logger.debug("[%s] using obonet cache at %s", prefix, obonet_json_gz_path)
        return from_obonet(get_gzipped_graph(obonet_json_gz_path))

    if has_nomenclature_plugin(prefix):
        obo = run_nomenclature_plugin(prefix, version=version)
        logger.debug("[%s] caching nomenclature plugin", prefix)
        obo.write_default(force=rewrite)
        return obo

    logger.debug("[%s] no obonet cache found at %s", prefix, obonet_json_gz_path)

    ontology_format, path = _ensure_ontology_path(prefix, force=force, version=version)
    if path is None:
        raise NoBuild
    elif ontology_format == "obo":
        pass  # all gucci
    elif ontology_format == "owl":
        from bioontologies import robot

        _converted_obo_path = path.with_suffix(".obo")
        if prefix in REQUIRES_NO_ROBOT_CHECK:
            robot_check = False
        robot.convert(path, _converted_obo_path, check=robot_check)
        path = _converted_obo_path
    else:
        raise UnhandledFormat(f"[{prefix}] unhandled ontology file format: {path.suffix}")

    from .reader import from_obo_path

    obo = from_obo_path(path, prefix=prefix, strict=strict)
    if version is not None:
        if obo.data_version is None:
            logger.warning("[%s] did not have a version, overriding with %s", obo.ontology, version)
            obo.data_version = version
        elif obo.data_version != version:
            logger.warning(
                "[%s] had version %s, overriding with %s", obo.ontology, obo.data_version, version
            )
            obo.data_version = version
    obo.write_default(force=rewrite)
    return obo


def _ensure_ontology_path(
    prefix: str, force, version
) -> Union[Tuple[str, Path], Tuple[None, None]]:
    for ontology_format, url in [  # noqa:B007
        ("obo", bioregistry.get_obo_download(prefix)),
        ("owl", bioregistry.get_owl_download(prefix)),
        ("json", bioregistry.get_json_download(prefix)),
    ]:
        if url is not None:
            try:
                path = Path(ensure_path(prefix, url=url, force=force, version=version))
            except urllib.error.HTTPError:
                continue
            else:
                return ontology_format, path
    return None, None


#: Obonet/Pronto can't parse these (consider converting to OBO with ROBOT?)
CANT_PARSE = {
    "agro",
    "aro",
    "bco",
    "caro",
    "cco",
    "chmo",
    "cido",
    "covoc",
    "cto",
    "cvdo",
    "dicom",
    "dinto",
    "emap",
    "epso",
    "eupath",
    "fbbi",
    "fma",
    "fobi",
    "foodon",
    "genepio",
    "hancestro",
    "hom",
    "hso",
    "htn",  # Unknown string format: creation: 16MAY2017
    "ico",
    "idocovid19",
    "labo",
    "mamo",
    "mfmo",
    "mfo",
    "mfomd",
    "miapa",
    "mo",
    "oae",
    "ogms",  # Unknown string format: creation: 16MAY2017
    "ohd",
    "ons",
    "oostt",
    "opmi",
    "ornaseq",
    "orth",
    "pdro",
    "probonto",
    "psdo",
    "reo",
    "rex",
    "rnao",
    "sepio",
    "sio",
    "spd",
    "sweetrealm",
    "txpo",
    "vido",
    "vt",
    "xl",
}
SKIP = {
    "ncbigene",  # too big, refs acquired from other dbs
    "pubchem.compound",  # to big, can't deal with this now
    "gaz",  # Gazetteer is irrelevant for biology
    "ma",  # yanked
    "bila",  # yanked
    # FIXME below
    "emapa",  # recently changed with EMAP... not sure what the difference is anymore
    "kegg.genes",
    "kegg.genome",
    "kegg.pathway",
    # URL is wrong
    "ensemblglossary",
    # Too much junk
    "biolink",
}

X = TypeVar("X")


def iter_helper(
    f: Callable[[str], Mapping[str, X]],
    leave: bool = False,
    strict: bool = True,
    **kwargs,
) -> Iterable[Tuple[str, str, X]]:
    """Yield all mappings extracted from each database given."""
    for prefix, mapping in iter_helper_helper(f, strict=strict, **kwargs):
        it = tqdm(
            mapping.items(),
            desc=f"iterating {prefix}",
            leave=leave,
            unit_scale=True,
            disable=None,
        )
        for key, value in it:
            value = value.strip('"').replace("\n", " ").replace("\t", " ").replace("  ", " ")
            if value:
                yield prefix, key, value


def _prefixes(
    skip_below: Optional[str] = None,
    skip_below_inclusive: bool = True,
    skip_pyobo: bool = False,
    skip_set: Optional[Set[str]] = None,
) -> Iterable[str]:
    for prefix, resource in sorted(bioregistry.read_registry().items()):
        if resource.no_own_terms:
            continue
        if prefix in SKIP:
            tqdm.write(f"skipping {prefix} because in default skip set")
            continue
        if skip_set and prefix in skip_set:
            tqdm.write(f"skipping {prefix} because in skip set")
            continue
        if skip_below is not None:
            if skip_below_inclusive:
                if prefix < skip_below:
                    continue
            else:
                if prefix <= skip_below:
                    continue
        has_pyobo = has_nomenclature_plugin(prefix)
        has_download = resource.has_download()
        if skip_pyobo and has_pyobo:
            continue
        if not has_pyobo and not has_download:
            continue
        yield prefix


def iter_helper_helper(
    f: Callable[[str], X],
    use_tqdm: bool = True,
    skip_below: Optional[str] = None,
    skip_below_inclusive: bool = True,
    skip_pyobo: bool = False,
    skip_set: Optional[Set[str]] = None,
    strict: bool = True,
    **kwargs,
) -> Iterable[Tuple[str, X]]:
    """Yield all mappings extracted from each database given.

    :param f: A function that takes a prefix and gives back something that will be used by an outer function.
    :param use_tqdm: If true, use the tqdm progress bar
    :param skip_below: If true, skip sources whose names are less than this (used for iterative curation
    :param skip_pyobo: If true, skip sources implemented in PyOBO
    :param skip_set: A pre-defined blacklist to skip
    :param strict: If true, will raise exceptions and crash the program instead of logging them.
    :param kwargs: Keyword arguments passed to ``f``.
    :yields: A prefix and the result of the callable ``f``

    :raises TypeError: If a type error is raised, it gets re-raised
    :raises urllib.error.HTTPError: If the resource could not be downloaded
    :raises urllib.error.URLError: If another problem was encountered during download
    :raises ValueError: If the data was not in the format that was expected (e.g., OWL)
    """
    prefixes = list(
        _prefixes(
            skip_set=skip_set,
            skip_below=skip_below,
            skip_pyobo=skip_pyobo,
            skip_below_inclusive=skip_below_inclusive,
        )
    )
    prefix_it = tqdm(
        prefixes, disable=not use_tqdm, desc=f"Building with {f.__name__}()", unit="resource"
    )
    for prefix in prefix_it:
        prefix_it.set_postfix(prefix=prefix)
        try:
            yv = f(prefix, **kwargs)  # type:ignore
        except urllib.error.HTTPError as e:
            logger.warning("[%s] HTTP %s: unable to download %s", prefix, e.getcode(), e.geturl())
            if strict and not bioregistry.is_deprecated(prefix):
                raise
        except urllib.error.URLError:
            logger.warning("[%s] unable to download", prefix)
            if strict and not bioregistry.is_deprecated(prefix):
                raise
        except MissingPrefix as e:
            logger.warning("[%s] missing prefix: %s", prefix, e)
            if strict and not bioregistry.is_deprecated(prefix):
                raise e
        except subprocess.CalledProcessError:
            logger.warning("[%s] ROBOT was unable to convert OWL to OBO", prefix)
        except UnhandledFormat as e:
            logger.warning("[%s] %s", prefix, e)
        except ValueError as e:
            if _is_xml(e):
                # this means that it tried doing parsing on an xml page
                logger.info(
                    "no resource available for %s. See http://www.obofoundry.org/ontology/%s",
                    prefix,
                    prefix,
                )
            else:
                logger.exception(
                    "[%s] got exception %s while parsing", prefix, e.__class__.__name__
                )
        except TypeError as e:
            logger.exception("[%s] got exception %s while parsing", prefix, e.__class__.__name__)
            if strict:
                raise e
        else:
            yield prefix, yv


def _is_xml(e) -> bool:
    return str(e).startswith("Tag-value pair parsing failed for:") or str(e).startswith(
        'Tag-value pair parsing failed for:\n<?xml version="1.0" encoding="UTF-8"?>'
    )


def _prep_dir(directory: Union[None, str, pathlib.Path]) -> pathlib.Path:
    if directory is None:
        rv = DATABASE_DIRECTORY
    elif isinstance(directory, str):
        rv = pathlib.Path(directory)
    elif isinstance(directory, pathlib.Path):
        rv = directory
    else:
        raise TypeError
    rv.mkdir(parents=True, exist_ok=True)
    return rv


def db_output_helper(
    f: Callable[..., Iterable[Tuple[str, ...]]],
    db_name: str,
    columns: Sequence[str],
    *,
    directory: Union[None, str, pathlib.Path] = None,
    strict: bool = True,
    use_gzip: bool = True,
    summary_detailed: Optional[Sequence[int]] = None,
    **kwargs,
) -> List[pathlib.Path]:
    """Help output database builds.

    :param f: A function that takes a prefix and gives back something that will be used by an outer function.
    :param db_name: name of the output resource (e.g., "alts", "names")
    :param columns: The names of the columns
    :param directory: The directory to output everything, or defaults to :data:`pyobo.constants.DATABASE_DIRECTORY`.
    :param strict: Passed to ``f`` by keyword
    :param kwargs: Passed to ``f`` by splat
    :returns: A sequence of paths that got created.
    """
    directory = _prep_dir(directory)

    c: typing.Counter[str] = Counter()
    c_detailed: typing.Counter[Tuple[str, ...]] = Counter()

    if use_gzip:
        db_path = directory.joinpath(f"{db_name}.tsv.gz")
    else:
        db_path = directory.joinpath(f"{db_name}.tsv")
    db_sample_path = directory.joinpath(f"{db_name}_sample.tsv")
    db_summary_path = directory.joinpath(f"{db_name}_summary.tsv")
    db_summary_detailed_path = directory.joinpath(f"{db_name}_summary_detailed.tsv")

    logger.info("writing %s to %s", db_name, db_path)
    logger.info("writing %s sample to %s", db_name, db_sample_path)
    it = f(strict=strict, **kwargs)
    with gzip.open(db_path, mode="wt") if use_gzip else open(db_path, "w") as gzipped_file:
        writer = get_writer(gzipped_file)

        # for the first 10 rows, put it in a sample file too
        with open(db_sample_path, "w") as sample_file:
            sample_writer = get_writer(sample_file)

            # write header
            writer.writerow(columns)
            sample_writer.writerow(columns)

            for row, _ in zip(it, range(10)):
                c[row[0]] += 1
                if summary_detailed is not None:
                    c_detailed[tuple(row[i] for i in summary_detailed)] += 1
                writer.writerow(row)
                sample_writer.writerow(row)

        # continue just in the gzipped one
        for row in it:
            c[row[0]] += 1
            if summary_detailed is not None:
                c_detailed[tuple(row[i] for i in summary_detailed)] += 1
            writer.writerow(row)

    logger.info(f"writing {db_name} summary to {db_summary_path}")
    with open(db_summary_path, "w") as file:
        writer = get_writer(file)
        writer.writerows(c.most_common())

    if summary_detailed is not None:
        logger.info(f"writing {db_name} detailed summary to {db_summary_detailed_path}")
        with open(db_summary_detailed_path, "w") as file:
            writer = get_writer(file)
            writer.writerows((*keys, v) for keys, v in c_detailed.most_common())

    db_metadata_path = directory.joinpath(f"{db_name}_metadata.json")
    with open(db_metadata_path, "w") as file:
        json.dump(
            {
                "version": get_version(),
                "git_hash": get_git_hash(),
                "date": datetime.datetime.now().strftime("%Y-%m-%d-%H-%M"),
                "count": sum(c.values()),
            },
            file,
            indent=2,
        )

    rv: List[pathlib.Path] = [
        db_metadata_path,
        db_path,
        db_sample_path,
        db_summary_path,
    ]
    if summary_detailed:
        rv.append(db_summary_detailed_path)
    return rv



# -*- coding: utf-8 -*-

"""Convert ExPASy to OBO."""

import logging
from collections import defaultdict
from typing import Dict, Iterable, Mapping, Optional, Set, Tuple

from .utils import get_go_mapping
from ..struct import Obo, Reference, Synonym, Term
from ..struct.typedef import enables, has_member
from ..utils.path import ensure_path

__all__ = [
    "ExpasyGetter",
]

PREFIX = "eccode"
EXPASY_DATABASE_URL = "ftp://ftp.expasy.org/databases/enzyme/enzyme.dat"
EXPASY_TREE_URL = "ftp://ftp.expasy.org/databases/enzyme/enzclass.txt"

logger = logging.getLogger(__name__)

#: The identifier of the entry (One)
ID = "ID"
#: Description (One)
DE = "DE"
#: Additional names/synonyms (Many)
AN = "AN"
#: Chemical Reaction String (One)
CA = "CA"
#: Comments (One - consider as free text)
CC = "CC"
#: List of cofactors? (Many)
CF = "CF"
#: ProSite Identifier (optional) (Many)
PR = "PR"
#: Reference to UniProt or SwissProt (Many)
DR = "DR"


class ExpasyGetter(Obo):
    """A getter for ExPASy Enzyme Classes."""

    bioversions_key = ontology = PREFIX
    typedefs = [has_member, enables]
    root_terms = [
        Reference(prefix="eccode", identifier="1"),
        Reference(prefix="eccode", identifier="2"),
        Reference(prefix="eccode", identifier="3"),
        Reference(prefix="eccode", identifier="4"),
        Reference(prefix="eccode", identifier="5"),
        Reference(prefix="eccode", identifier="6"),
        Reference(prefix="eccode", identifier="7"),
    ]
    idspaces = {
        "uniprot": "https://bioregistry.io/uniprot:",
        "eccode": "https://bioregistry.io/eccode:",
        "GO": "http://purl.obolibrary.org/obo/GO_",
        "RO": "http://purl.obolibrary.org/obo/RO_",
    }

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get ExPASy as OBO."""
    return ExpasyGetter(force=force)


def get_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get the ExPASy terms."""
    tree_path = ensure_path(PREFIX, url=EXPASY_TREE_URL, version=version, force=force)
    with open(tree_path) as file:
        tree = get_tree(file)

    terms: Dict[str, Term] = {}
    child_to_parents = defaultdict(list)
    for ec_code, data in tree.items():
        terms[ec_code] = Term(
            reference=Reference(prefix=PREFIX, identifier=ec_code, name=data["name"]),
        )
        for child_data in data.get("children", []):
            child_ec_code = child_data["identifier"]
            child_to_parents[child_ec_code].append(ec_code)

    for child_ec_code, parents_ec_codes in child_to_parents.items():
        terms[child_ec_code].parents = [
            terms[parent_ec_code].reference for parent_ec_code in parents_ec_codes
        ]

    database_path = ensure_path(PREFIX, url=EXPASY_DATABASE_URL, version=version)
    with open(database_path) as file:
        _data = get_database(file)

    ec2go = get_ec2go(version=version)

    ec_code_to_alt_ids = {}
    for ec_code, data in _data.items():
        parent_ec_code = data["parent"]["identifier"]
        parent_term = terms[parent_ec_code]

        synonyms = [Synonym(name=synonym) for synonym in data.get("synonyms", [])]
        if data["alt_ids"]:
            alt_ids = data["alt_ids"][0].rstrip(".")
            if "and" not in alt_ids:
                ec_code_to_alt_ids[ec_code] = [
                    alt_ids,
                ]
            else:
                ec_code_to_alt_ids[ec_code] = [
                    alt_id.rstrip(",") for alt_id in alt_ids.split(" ") if alt_id != "and"
                ]

        concept = data["concept"]
        try:
            name = concept["name"]
        except KeyError:
            continue
            # raise

        term = terms[ec_code] = Term(
            reference=Reference(prefix=PREFIX, identifier=ec_code, name=name),
            parents=[parent_term.reference],
            synonyms=synonyms,
        )
        for domain in data.get("domains", []):
            term.append_relationship(
                has_member,
                Reference(prefix=domain["namespace"], identifier=domain["identifier"]),
            )
        for protein in data.get("proteins", []):
            term.append_relationship(
                has_member,
                Reference(
                    prefix=protein["namespace"],
                    identifier=protein["identifier"],
                    name=protein["name"],
                ),
            )
        for go_id, go_name in ec2go.get(ec_code, []):
            term.append_relationship(
                enables, Reference(prefix="GO", identifier=go_id, name=go_name)
            )

    return terms.values()


"""TREE"""


def normalize_expasy_id(expasy_id: str) -> str:
    """Return a standardized ExPASy identifier string.

    :param expasy_id: A possibly non-normalized ExPASy identifier
    """
    return expasy_id.replace(" ", "")


def give_edge(unnormalized_ec_code: str) -> Tuple[int, Optional[str], str]:
    """Return a (parent, child) tuple for given id."""
    levels = [x for x in unnormalized_ec_code.replace(" ", "").replace("-", "").split(".") if x]
    level = len(levels)

    if level == 1:
        parent_id = None
    else:
        parent_id = ".".join(levels[:-1])

    return level, parent_id, ".".join(levels)


def get_tree(lines: Iterable[str]):
    """Get the ExPASy tree mapping."""
    rv = {}
    for line in lines:
        if not line[0].isnumeric():
            continue
        level, parent_expasy_id, expasy_id = give_edge(line[:7])
        name = line[11:]
        name = name.strip().strip(".")

        rv[expasy_id] = {
            "concept": {
                "namespace": PREFIX,
                "identifier": expasy_id,
            },
            "name": name,
            "level": level,
            "children": [],
        }
        if parent_expasy_id is not None:
            rv[expasy_id]["parent"] = {
                "namespace": PREFIX,
                "identifier": parent_expasy_id,
            }
            rv[parent_expasy_id]["children"].append(rv[expasy_id]["concept"])  # type:ignore

    return rv


def get_database(lines: Iterable[str]) -> Mapping:
    """Parse the ExPASy database file and returns a list of enzyme entry dictionaries.

    :param lines: An iterator over the ExPASy database file or file-like
    """
    rv = {}
    for groups in _group_by_id(lines):
        _, expasy_id = groups[0]

        rv[expasy_id] = ec_data_entry = {
            "concept": {
                "namespace": PREFIX,
                "identifier": expasy_id,
            },
            "parent": {
                "namespace": PREFIX,
                "identifier": expasy_id.rsplit(".", 1)[0],
            },
            "synonyms": [],
            "cofactors": [],
            "domains": [],
            "proteins": [],
            "alt_ids": [],
        }

        for descriptor, value in groups[1:]:
            if descriptor == "//":
                continue
            elif descriptor == DE and value == "Deleted entry.":
                continue
            elif descriptor == DE and value.startswith("Transferred entry: "):
                value = value[len("Transferred entry: ") :].rstrip()
                ec_data_entry["transfer_id"] = value
            elif descriptor == DE:
                ec_data_entry["concept"]["name"] = value.rstrip(".")  # type:ignore
            elif descriptor == AN:
                ec_data_entry["synonyms"].append(value.rstrip("."))  # type:ignore
            elif descriptor == PR:
                value = value[len("PROSITE; ") : -1]  # remove trailing comma
                ec_data_entry["domains"].append(  # type:ignore
                    {
                        "namespace": "prosite",
                        "identifier": value,
                    }
                )
            elif descriptor == DR:
                for uniprot_entry in value.replace(" ", "").split(";"):
                    if not uniprot_entry:
                        continue
                    uniprot_id, uniprot_accession = uniprot_entry.split(",")
                    ec_data_entry["proteins"].append(  # type:ignore
                        dict(
                            namespace="uniprot",
                            name=uniprot_accession,
                            identifier=uniprot_id,
                        )
                    )

    for expasy_id, data in rv.items():
        transfer_id = data.pop("transfer_id", None)
        if transfer_id is not None:
            rv[expasy_id]["alt_ids"].append(transfer_id)  # type:ignore

    return rv


def _group_by_id(lines):
    """Group lines by identifier."""
    groups = []
    for line in lines:  # TODO replace with itertools.groupby
        line = line.strip()

        if line.startswith("ID"):
            groups.append([])

        if not groups:
            continue

        descriptor = line[:2]
        value = line[5:]

        groups[-1].append((descriptor, value))

    return groups


def get_ec2go(version: str) -> Mapping[str, Set[Tuple[str, str]]]:
    """Get the EC mapping to GO activities."""
    url = "http://current.geneontology.org/ontology/external2go/ec2go"
    path = ensure_path(PREFIX, url=url, name="ec2go.tsv", version=version)
    return get_go_mapping(path, "EC")


if __name__ == "__main__":
    ExpasyGetter.cli()



# -*- coding: utf-8 -*-

"""Converter for miRBase."""

import gzip
import logging
from typing import Iterable, List, Mapping

from tqdm.auto import tqdm

from ..struct import Obo, Reference, Synonym, Term, from_species
from ..struct.typedef import has_mature
from ..utils.cache import cached_mapping
from ..utils.path import ensure_df, ensure_path, prefix_directory_join

__all__ = [
    "MiRBaseGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "mirbase"
MIRBASE_MATURE_PREFIX = "mirbase.mature"
BASE_URL = "https://www.mirbase.org/ftp"

xref_mapping = {
    "entrezgene": "ncbigene",
    "targets:pictar-vert": "pictar",
    "targets:mirte": "mirte",
    "targets:pictar-fly": "pictar",
}


class MiRBaseGetter(Obo):
    """An ontology representation of miRBase's miRNA nomenclature."""

    ontology = bioversions_key = PREFIX
    typedefs = [from_species, has_mature]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get miRBase as OBO."""
    return MiRBaseGetter(force=force)


def get_terms(version: str, force: bool = False) -> List[Term]:
    """Parse miRNA data from filepath and convert it to dictionary."""
    url = f"{BASE_URL}/{version}/miRNA.dat.gz"
    definitions_path = ensure_path(PREFIX, url=url, version=version, force=force)

    file_handle = (
        gzip.open(definitions_path, "rt")
        if definitions_path.endswith(".gz")
        else open(definitions_path)
    )
    with file_handle as file:
        return list(_process_definitions_lines(file, version=version, force=force))


def _prepare_organisms(version: str, force: bool = False):
    url = f"{BASE_URL}/{version}/organisms.txt.gz"
    df = ensure_df(PREFIX, url=url, sep="\t", dtype={"#NCBI-taxid": str}, version=version)
    return {division: (taxonomy_id, name) for _, division, name, _tree, taxonomy_id in df.values}


def _prepare_aliases(version: str, force: bool = False) -> Mapping[str, List[str]]:
    url = f"{BASE_URL}/{version}/aliases.txt.gz"
    df = ensure_df(PREFIX, url=url, sep="\t", version=version)
    return {
        mirbase_id: [s.strip() for s in synonyms.split(";") if s and s.strip()]
        for mirbase_id, synonyms in df.values
    }


def _process_definitions_lines(
    lines: Iterable[str], version: str, force: bool = False
) -> Iterable[Term]:
    """Process the lines of the definitions file."""
    organisms = _prepare_organisms(version, force=force)
    aliases = _prepare_aliases(version, force=force)

    groups: List[List[str]] = []

    for line in lines:  # TODO replace with itertools.groupby
        if line.startswith("ID"):
            groups.append([])
        groups[-1].append(line)

    for group in tqdm(groups, desc=f"mapping {PREFIX}"):
        name = group[0][5:23].strip()
        qualifier, dtype, species_code, length = map(
            str.strip, group[0][23:].strip().rstrip(".").split(";")
        )
        identifier = group[2][3:-2].strip()
        definition = group[4][3:-1].strip()

        synonyms = [Synonym(name=alias) for alias in aliases.get(identifier, []) if alias != name]
        mature_mirna_lines = [i for i, element in enumerate(group) if "FT   miRNA    " in element]

        matures = []
        for index in mature_mirna_lines:
            # location = group[index][10:-1].strip()
            accession = group[index + 1][33:-2]
            product = group[index + 2][31:-2]
            product_reference = Reference(
                prefix=MIRBASE_MATURE_PREFIX,
                identifier=accession,
                name=product,
            )
            if product.endswith("3p") or product.endswith("5p"):
                matures.append(product_reference)
            else:
                pass
                # logger.warning(f'Whats going on {group[index]}')

        xrefs = []
        for line in group:
            if not line.startswith("DR"):
                continue
            line = line[len("DR   ") :].strip().rstrip(".")
            xref_prefix, xref_identifier, xref_label = map(str.strip, line.split(";"))
            xref_prefix = xref_prefix.lower()
            xref_prefix = xref_mapping.get(xref_prefix, xref_prefix)
            xrefs.append(
                Reference(prefix=xref_prefix, identifier=xref_identifier, name=xref_label or None)
            )

        # TODO add pubmed references

        term = Term(
            reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
            definition=definition,
            xrefs=xrefs,
            synonyms=synonyms,
        )

        species_identifier, species_name = organisms[species_code]
        term.set_species(species_identifier, species_name)
        term.extend_relationship(has_mature, matures)

        yield term


def get_mature_to_premature(version: str) -> Mapping[str, str]:
    """Get a mapping from mature miRNAs to their parents."""

    @cached_mapping(
        path=prefix_directory_join(
            PREFIX, name=f"{PREFIX}.mature_to_{PREFIX}.tsv", version=version
        ),
        header=["mirbase.mature_id", "mirbase_id"],
    )
    def _inner():
        return {
            mature.identifier: term.identifier
            for term in get_terms(version)
            for mature in term.get_relationships(has_mature)
        }

    return _inner()


def get_mature_id_to_name(version: str) -> Mapping[str, str]:
    """Get a mapping from mature miRNAs to their parents."""

    @cached_mapping(
        path=prefix_directory_join(PREFIX, name=f"{PREFIX}.mature_mapping.tsv", version=version),
        header=["mirbase.mature_id", "name"],
    )
    def _inner():
        return {
            mature.identifier: mature.name
            for term in get_terms(version)
            for mature in term.get_relationships(has_mature)
        }

    return _inner()


if __name__ == "__main__":
    get_obo(force=True).write_default(force=True, write_obograph=True, write_obo=True)



# -*- coding: utf-8 -*-

"""CLI for PyOBO."""

import logging
import os
import sys
from operator import itemgetter

import click
import humanize
from more_click import verbose_option
from tabulate import tabulate

from .aws import main as aws_main
from .database import main as database_main
from .lookup import lookup
from ..apps.cli import main as apps_main
from ..constants import RAW_DIRECTORY
from ..plugins import has_nomenclature_plugin, iter_nomenclature_plugins
from ..registries import iter_cached_obo
from ..utils.io import get_writer
from ..xrefdb.canonicalizer import Canonicalizer, get_priority_curie, remap_file_stream
from ..xrefdb.priority import DEFAULT_PRIORITY_LIST

__all__ = ["main"]

logger = logging.getLogger(__name__)


@click.group()
@click.version_option()
def main():
    """CLI for PyOBO."""


_ORDERING_TEXT = ", ".join(f"{i}) {x}" for i, x in enumerate(DEFAULT_PRIORITY_LIST, start=1))


@main.command(help=f"Prioritize a CURIE from ordering: {_ORDERING_TEXT}")
@click.argument("curie")
def prioritize(curie: str):
    """Prioritize a CURIE."""
    priority_curie = get_priority_curie(curie)
    click.secho(priority_curie)


@main.command()
@click.option("-i", "--file-in", type=click.File("r"), default=sys.stdin)
@click.option("-o", "--file-out", type=click.File("w"), default=sys.stdout)
@click.option("--column", type=int, default=0, show_default=True)
@click.option("--sep", default="\t", show_default=True)
def recurify(file_in, file_out, column: int, sep: str):
    """Remap a column in a given file stream."""
    remap_file_stream(file_in=file_in, file_out=file_out, column=column, sep=sep)


@main.command()
@verbose_option
def cache():
    """Cache all resources."""
    for obo in iter_nomenclature_plugins():
        click.secho(f"Caching {obo.ontology}", bold=True, fg="green")
        obo.write_default()


@main.command()
@click.option("--remove-obo", is_flag=True)
def clean(remove_obo: bool):
    """Delete all cached files."""
    suffixes = [
        "_mappings.tsv",
        ".mapping.tsv",
        "_synonyms.tsv",
        ".obo.pickle",
        ".obo.json.gz",
        ".owl",
    ]
    if remove_obo:
        suffixes.append(".obo")
    for directory in os.listdir(RAW_DIRECTORY):
        d = os.path.join(RAW_DIRECTORY, directory)
        if not os.path.isdir(d):
            continue
        entities = list(os.listdir(d))
        if not entities:
            os.rmdir(d)
        obo_pickle = os.path.join(d, f"{directory}.obo.pickle")
        if os.path.exists(obo_pickle):
            os.remove(obo_pickle)
        for f in entities:
            if any(f.endswith(suffix) for suffix in suffixes):
                os.remove(os.path.join(d, f))


@main.command()
def ls():
    """List how big all of the OBO files are."""
    entries = [(prefix, os.path.getsize(path)) for prefix, path in iter_cached_obo()]
    entries = [
        (prefix, humanize.naturalsize(size), "✅" if not has_nomenclature_plugin(prefix) else "❌")
        for prefix, size in sorted(entries, key=itemgetter(1), reverse=True)
    ]
    click.echo(tabulate(entries, headers=["Source", "Size", "OBO"]))


@main.command()
@verbose_option
@click.option("-f", "--file", type=click.File("w"))
def remapping(file):
    """Make a canonical remapping."""
    canonicalizer = Canonicalizer.get_default()
    writer = get_writer(file)
    writer.writerow(["input", "canonical"])
    writer.writerows(canonicalizer.iterate_flat_mapping())


main.add_command(lookup)
main.add_command(apps_main)
main.add_command(aws_main)
main.add_command(database_main)

if __name__ == "__main__":
    main()
